# This file is part of MyPaint.
# Copyright (C) 2008-2013 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2008-2018 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


## Imports

from __future__ import division, print_function
import random
from math import floor, ceil, log, exp
import math
import weakref
import contextlib
import logging

from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GLib
import cairo
import numpy as np

from lib import helpers, tiledsurface, pixbufsurface
from lib.observable import event
import lib.layer
from . import cursor
from .drawutils import render_checks
import gui.style
import lib.color
from lib.pycompat import xrange

logger = logging.getLogger(__name__)

## Class definitions


class TiledDrawWidget (Gtk.EventBox):
    """Widget for showing a lib.document.Document

    Rendering is delegated to a dedicated class: see `CanvasRenderer`.

    """

    ## Register a GType name for Glade, GtkBuilder etc.
    __gtype_name__ = "TiledDrawWidget"

    # List of weakrefs to all known TDWs.
    __tdw_refs = []

    @classmethod
    def get_active_tdw(kin):  # noqa: N804
        """Returns the most recently created or entered TDW.
        """
        # Find and return the first visible, mapped etc. TDW in the list
        invis_refs = []
        active_tdw = None
        while len(kin.__tdw_refs) > 0:
            tdw_ref = kin.__tdw_refs[0]
            tdw = tdw_ref()
            if tdw is not None:
                if tdw.get_window() is not None and tdw.get_visible():
                    active_tdw = tdw
                    break
                else:
                    invis_refs.append(tdw_ref)
            kin.__tdw_refs.pop(0)
        kin.__tdw_refs.extend(invis_refs)
        assert active_tdw is not None
        return active_tdw

    @classmethod
    def get_visible_tdws(cls):
        """Iterates across all mapped and visible TDWs"""
        for tdw_ref in cls.__tdw_refs:
            tdw = tdw_ref()
            if tdw and tdw.get_window() and tdw.get_visible():
                yield tdw

    @classmethod
    def get_tdw_under_device(cls, device):
        """Get the TDW directly under a device's pointer

        :param Gdk.Device device: the device to look up
        :rtype: tuple
        :returns: (tdw, x, y)

        This classmethod returns
        the TDW directly under the master pointer for a given device,
        and the pointer's coordinates
        relative to the top-left of that TDW's window.
        It is intended for use in picker code,
        and within the kinds of device-specific pointer grabs
        which the lib.picker presenters establish.
        Most pointer event handling code doesn't need to call this.

        If no known tdw is under the device's position,
        the tuple `(None, -1, -1)` is returned.

        """
        # get_last_event_window() does not work in pointer grabs.
        #  dev_win = device.get_last_event_window()
        # But we want *under* the pointer semantics anyway, and
        # get_window_at_position() behaves itself in grabs. Hopefully.
        dev_win, win_x, win_y = device.get_window_at_position()
        if dev_win is None:
            return (None, -1, -1)
        # The window returned is that of the tdw's renderer, i.e. that
        # of the tdw itself under the current working assumptions.
        for tdw in cls.get_visible_tdws():
            tdw_win = tdw.get_window()
            assert tdw_win is tdw.renderer.get_window(), (
                "Picking will break in this packing configuration "
                "because the tdw's event window is not the same as "
                "its descendent renderer's window."
            )
            if tdw_win is dev_win:
                return (tdw, win_x, win_y)
        return (None, -1, -1)

    def __init__(self, idle_redraw_priority=None):
        """Instantiate a TiledDrawWidget.

        """
        super(TiledDrawWidget, self).__init__()

        if __name__ == '__main__':
            app = None
        else:
            from . import application
            app = application.get_app()
        self.app = app
        self.doc = None

        self.add_events(
            Gdk.EventMask.BUTTON_PRESS_MASK |
            Gdk.EventMask.BUTTON_RELEASE_MASK |
            Gdk.EventMask.POINTER_MOTION_MASK
        )
        # Support smooth scrolling unless configured not to
        if app and app.preferences.get("ui.support_smooth_scrolling", True):
            self.add_events(Gdk.EventMask.SMOOTH_SCROLL_MASK)

        self.last_painting_pos = None

        self.renderer = CanvasRenderer(
            self,
            idle_redraw_priority = idle_redraw_priority,
        )
        self.add(self.renderer)
        self.renderer.update_cursor()  # get the initial cursor right

        self.add_events(Gdk.EventMask.ENTER_NOTIFY_MASK)
        self.connect("enter-notify-event", self.enter_notify_cb)
        self.__tdw_refs.insert(0, weakref.ref(self))

        self._last_alloc_pos = (0, 0)
        self.connect("size-allocate", self._size_allocate_cb)
        self.connect("realize", self._realize_cb)

        #: Scroll to match appearing/disappearing sidebars and toolbars.
        self.scroll_on_allocate = True

        forwarder = self._announce_transformation_updated
        self.renderer.transformation_updated += forwarder

    def _announce_transformation_updated(self, *args):
        self.transformation_updated()

    @event
    def transformation_updated(self):
        """Forwarded event: transformation was updated"""

    def _size_allocate_cb(self, widget, alloc):
        """Allow for allocation changes under certain circumstances

        We need to allow for changes like toolbars or sidebars appearing or
        disappearing on the top or the left.  The canvas position should remain
        stationary on the screen in these cases.  This size-allocate handler
        deals with that by issuing appropriate scroll() events.

        See also `scroll_on_allocate`.

        """
        # Capture the last allocated position in toplevel coords
        toplevel = self.get_toplevel()
        new_pos = self.translate_coordinates(toplevel, alloc.x, alloc.y)
        old_pos = self._last_alloc_pos
        self._last_alloc_pos = new_pos
        # When things change measurably, scroll to make up the difference
        if not self.scroll_on_allocate:
            return
        if None in (old_pos, new_pos):
            return
        if old_pos != new_pos:
            dx = new_pos[0] - old_pos[0]
            dy = new_pos[1] - old_pos[1]
            self.renderer.scroll(dx, dy, ongoing=False)

    def _realize_cb(self, widget):
        logger.debug("Turning off event compression for %r's window", widget)
        win = widget.get_window()
        win.set_event_compression(False)

    def set_model(self, model):
        assert self.doc is None
        renderer = self.renderer
        model.canvas_area_modified += renderer.canvas_modified_cb
        root = model.layer_stack
        root.current_path_updated += renderer.current_layer_changed_cb
        root.layer_properties_changed += renderer.layer_props_changed_cb
        model.brush.brushinfo.observers.append(renderer.brush_modified_cb)
        model.frame_enabled_changed += renderer.frame_enabled_changed_cb
        model.frame_updated += renderer.frame_updated_cb
        self.doc = model
        self.renderer.queue_draw()

    def enter_notify_cb(self, widget, event):
        # Track the active TDW
        self_ref = weakref.ref(self)
        self.__tdw_refs.remove(self_ref)
        self.__tdw_refs.insert(0, self_ref)
        # Ensure the cursor reflects layer-locking changes etc.
        self.renderer.update_cursor()

    # Forward public API to delegates

    @property
    def scale(self):
        return self.renderer.scale

    @scale.setter
    def scale(self, n):
        self.renderer.scale = n

    @property
    def zoom_min(self):
        return self.renderer.zoom_min

    @zoom_min.setter
    def zoom_min(self, n):
        self.renderer.zoom_min = n

    @property
    def zoom_max(self):
        return self.renderer.zoom_max

    @zoom_max.setter
    def zoom_max(self, n):
        self.renderer.zoom_max = n

    @property
    def rotation(self):
        return self.renderer.rotation

    @property
    def mirrored(self):
        return self.renderer.mirrored

    @property
    def pixelize_threshold(self):
        return self.renderer.pixelize_threshold

    @pixelize_threshold.setter
    def pixelize_threshold(self, n):
        self.renderer.pixelize_threshold = n

    @property
    def display_overlays(self):
        return self.renderer.display_overlays

    @property
    def model_overlays(self):
        return self.renderer.model_overlays

    @property
    def overlay_layer(self):
        return self.renderer.overlay_layer

    @overlay_layer.setter
    def overlay_layer(self, l):
        self.renderer.overlay_layer = l

    @property
    def recenter_document(self):
        return self.renderer.recenter_document

    @property
    def display_to_model(self):
        return self.renderer.display_to_model

    @property
    def model_to_display(self):
        return self.renderer.model_to_display

    @property
    def set_override_cursor(self):
        return self.renderer.set_override_cursor

    def get_last_painting_pos(self):
        return self.last_painting_pos

    def set_last_painting_pos(self, value):
        self.last_painting_pos = value

    def get_pointer_in_model_coordinates(self):
        """Returns the pointer/cursor location in model coords.

        :returns: core pointer's position, as a model-relative (x, y)
        :rtype: tuple

        This should only be used in action callbacks, never for events.

        """
        win = self.get_window()
        display = self.get_display()
        devmgr = display and display.get_device_manager() or None
        coredev = devmgr and devmgr.get_client_pointer() or None
        if coredev and win:
            win_, x, y, mods = win.get_device_position_double(coredev)
            return self.display_to_model(x, y)
        else:
            return (0., 0.)

    @property
    def scroll(self):
        return self.renderer.scroll

    @property
    def get_center(self):
        return self.renderer.get_center

    @property
    def get_center_model_coords(self):
        return self.renderer.get_center_model_coords

    def get_corners_model_coords(self):
        """Returns the viewport corners in model coordinates.

        :returns: Corners [TL, TR, BR, BL] as (x, y) pairs of floats.
        :rtype: list

        See also lib.helpers.rotated_rectangle_bbox() if you need to
        turn this into a bounding box in model-space.

        """
        alloc = self.get_allocation()
        x = alloc.x
        y = alloc.y
        w = alloc.width
        h = alloc.height
        corners = [(x, y), (x + w, y), (x + w, y + h), (x, y + h)]
        corners = [self.display_to_model(*p) for p in corners]
        return corners

    @property
    def recenter_on_model_coords(self):
        return self.renderer.recenter_on_model_coords

    @property
    def pick_color(self):
        return self.renderer.pick_color

    @property
    def queue_draw_area(self):
        return self.renderer.queue_draw_area

    # Transform logic

    @contextlib.contextmanager
    def _fixed_center(self, center=None, ongoing=True):
        """Keep a fixed center when zoom or rotation changes

        :param tuple center: Center of the rotation, display (X, Y)
        :param bool ongoing: Hint that this is in an ongoing change

        This context manager's cleanup phase applies a corrective
        transform which keeps the specified center in the same position
        on the screen. If the center isn't specified, the center pixel
        of the widget itself is used.

        It also queues a redraw.

        """
        # Determine the requested (or default) center in model space
        cx = None
        cy = None
        if center is not None:
            cx, cy = center
        if cx is None or cy is None:
            cx, cy = self.renderer.get_center()
        cx_model, cy_model = self.renderer.display_to_model(cx, cy)
        # Execute the changes in the body of the with statement
        yield
        # Corrective transform (& limits)
        self.renderer.scale = helpers.clamp(self.renderer.scale,
                                            self.renderer.zoom_min,
                                            self.renderer.zoom_max)
        cx_new, cy_new = self.renderer.model_to_display(cx_model, cy_model)
        self.renderer.translation_x += cx - cx_new
        self.renderer.translation_y += cy - cy_new
        # Redraw handling
        if ongoing:
            self.renderer.defer_hq_rendering()
        self.renderer.queue_draw()

    def zoom(self, zoom_step, center=None, ongoing=True):
        """Multiply the current scale factor"""
        with self._fixed_center(center, ongoing):
            self.renderer.scale *= zoom_step

    def set_zoom(self, zoom, center=None, ongoing=False):
        """Set the zoom to an exact value"""
        with self._fixed_center(center, ongoing):
            self.renderer.scale = zoom
        self.renderer.update_cursor()

    def rotate(self, angle_step, center=None, ongoing=True):
        """Rotate the view by a step"""
        if self.renderer.mirrored:
            angle_step = -angle_step
        with self._fixed_center(center, ongoing):
            self.renderer.rotation += angle_step

    def set_rotation(self, angle, ongoing=False):
        """Set the rotation to an exact value"""
        if self.renderer.mirrored:
            angle = -angle
        with self._fixed_center(None, ongoing):
            self.renderer.rotation = angle

    def mirror(self):
        with self._fixed_center(None, False):
            self.renderer.mirrored = not self.renderer.mirrored

    def set_mirrored(self, mirrored):
        """Set mirroring to a discrete state"""
        with self._fixed_center(None, False):
            self.renderer.mirrored = mirrored

    def get_transformation(self):
        """Returns a snapshot/memento/record of the current transformation.

        :rtype: a CanvasTransformation initialized with a copy of the current
          transformation variables.

        """
        tr = CanvasTransformation()
        tr.translation_x = self.renderer.translation_x
        tr.translation_y = self.renderer.translation_y
        tr.scale = self.renderer.scale
        tr.rotation = self.renderer.rotation
        tr.mirrored = self.renderer.mirrored
        return tr

    def set_transformation(self, transformation):
        """Sets the current transformation, and redraws.

        :param transformation: a CanvasTransformation object.

        """
        self.renderer.translation_x = transformation.translation_x
        self.renderer.translation_y = transformation.translation_y
        self.renderer.scale = transformation.scale
        self.renderer.rotation = transformation.rotation
        self.renderer.mirrored = transformation.mirrored
        self.renderer.queue_draw()
        self.renderer.update_cursor()

    def get_move_cursor_name_for_edge(self, cursor_pos, edge_p1, edge_p2,
                                      tolerance=5, finite=True):
        """Get move cursor & detect hits on a line between two points

        :param tuple cursor_pos: cursor position, as display (x, y)
        :param tuple edge_p1: point on the edge, as model (x, y)
        :param tuple edge_p2: point on the edge, as model (x, y)
        :param int tolerance: slack for cursor pos., in display pixels
        :param bool finite: if false, the edge extends beyond p1, p2
        :returns: move direction cursor & distance from line (cursor, d)
        :rtype: tuple

        This can be used by special input modes when resizing objects on
        screen, for example frame edges and the symmetry axis.

        If the returned cursor isn't None, its value is the name of the
        most appropriate move cursor to use during the move. This method
        does not return a normal vector in model space; you'll have to
        calculate that for yourself.

        See also:
        * gui.cursor.Name (naming consts for cursors)

        """
        # Work in screen pixels only
        x0, y0 = cursor_pos
        x1, y1 = self.model_to_display(*edge_p1)
        x2, y2 = self.model_to_display(*edge_p2)
        dx = x2 - x1
        dy = y2 - y1
        edge_len = math.sqrt(dx**2 + dy**2)
        if edge_len <= 0:
            dist1 = math.hypot(x0 - x1, y0 - y1)
            dist2 = math.hypot(x0 - x2, y0 - y2)
            return (None, min(dist1, dist2))
        # Perpendicular distance from a line.
        two_triarea = abs((dy * x0) - (dx * y0) - (x1 * y2) + (x2 * y1))
        perp_dist = two_triarea / edge_len
        if perp_dist > tolerance:
            return (None, perp_dist)
        # Rough approximation here, but good enough for hit calculation.
        # Omit points too far away from the edge's centre point.
        if finite:
            xmid = (x1 + x2) / 2.0
            ymid = (y1 + y2) / 2.0
            dist_from_midpt = math.sqrt(((xmid - x0) ** 2) +
                                        ((ymid - y0) ** 2))
            if dist_from_midpt > ((edge_len / 2.0) + tolerance):
                return (None, perp_dist)
        # Cursor name by sector.
        # Aiming for a cursor that looks perpendicular to the line.
        # Ish.
        theta = math.atan2(dy, dx)
        c = cursor.get_move_cursor_name_for_angle(theta + (math.pi / 2))
        return (c, perp_dist)


class CanvasTransformation (object):
    """Record of a TiledDrawWidget's canvas (view) transformation.
    """

    translation_x = 0.0
    translation_y = 0.0
    scale = 1.0
    rotation = 0.0
    mirrored = False

    def __repr__(self):
        return "<%s dx=%0.3f dy=%0.3f scale=%0.3f rot=%0.3f%s>" % (
            self.__class__.__name__,
            self.translation_x, self.translation_y,
            self.scale, self.rotation,
            (self.mirrored and " mirrored" or "")
        )


class DrawCursorMixin(object):
    """Mixin for renderer widgets needing a managed drawing cursor.

    Required members: self.doc, self.scale, Gtk.Widget stuff.

    """

    def init_draw_cursor(self):
        """Initialize internal fields for DrawCursorMixin"""
        self._override_cursor = None
        self._first_map_cb_id = self.connect("map", self._first_map_cb)

    def _first_map_cb(self, widget, *a):
        """Updates the cursor on the first map"""
        assert self.get_window() is not None
        assert self.get_mapped()
        self.disconnect(self._first_map_cb_id)
        self._first_map_cb_id = None
        self.update_cursor()

    def update_cursor(self):
        # Callback for updating the cursor
        if not self.get_mapped():
            return
        window = self.get_window()
        app = self.app
        if window is None:
            logger.error("update_cursor: no window")
            return
        override_cursor = self._override_cursor
        layer = self.doc._layers.current
        if self._insensitive_state_content:
            c = None
        elif override_cursor is not None:
            c = override_cursor
        elif self.doc is None:
            logger.error("update_cursor: no document")
            return
        elif not layer.get_paintable():
            # Cursor to represent that one cannot draw.
            # Often a red circle with a diagonal bar through it.
            c = Gdk.Cursor.new_for_display(
                window.get_display(), Gdk.CursorType.CIRCLE)
        elif app is None:
            logger.error("update_cursor: no app")
            return
        # Last two cases only pertain to FreehandMode cursors.
        # XXX refactor: bad for separation of responsibilities, put the
        # special cases in the mode class.
        elif app.preferences.get("cursor.freehand.style", None) == 'crosshair':
            c = app.cursors.get_freehand_cursor()
        else:
            radius, style = self._get_cursor_info()
            c = cursor.get_brush_cursor(radius, style, self.app.preferences)
        window.set_cursor(c)

    def set_override_cursor(self, cursor):
        """Set a cursor which will always be used.

        Used by the color picker. The override cursor will be used regardless
        of the criteria update_cursor() normally uses. Pass None to let it
        choose normally again.
        """
        self._override_cursor = cursor
        GLib.idle_add(self.update_cursor)

    def _get_cursor_info(self):
        """Return factors determining the cursor size and shape.
        """
        b = self.doc.brush.brushinfo
        base_radius = exp(b.get_base_value('radius_logarithmic'))
        r = base_radius
        r += 2 * base_radius * b.get_base_value('offset_by_random')
        r *= self.scale
        r += 0.5
        if b.is_eraser():
            style = cursor.BRUSH_CURSOR_STYLE_ERASER
        elif b.is_alpha_locked():
            style = cursor.BRUSH_CURSOR_STYLE_LOCK_ALPHA
        elif b.is_colorize():
            style = cursor.BRUSH_CURSOR_STYLE_COLORIZE
        else:
            style = cursor.BRUSH_CURSOR_STYLE_NORMAL
        return (r, style)

    def brush_modified_cb(self, settings):
        """Handles brush modifications: set up by the main TDW.
        """
        if settings & set(['radius_logarithmic', 'offset_by_random',
                           'eraser', 'lock_alpha', 'colorize']):
            # Reducing the number of updates is probably a good idea
            self.update_cursor()


def calculate_transformation_matrix(scale, rotation,
                                    translation_x, translation_y,
                                    mirrored):
    scale = scale
    # check if scale is almost a power of two
    scale_log2 = log(scale, 2)
    scale_log2_rounded = round(scale_log2)
    if abs(scale_log2 - scale_log2_rounded) < 0.01:
        scale = 2.0 ** scale_log2_rounded

    # maybe we should check if rotation is almost a multiple of 90 degrees?

    matrix = cairo.Matrix()
    matrix.translate(translation_x, translation_y)
    matrix.rotate(rotation)
    matrix.scale(scale, scale)

    # Align the translation such that (0,0) maps to an integer
    # screen pixel, to keep image rendering fast and sharp.
    x, y = matrix.transform_point(0, 0)
    inverse = cairo.Matrix(*list(matrix))
    assert not inverse.invert()
    x, y = inverse.transform_point(round(x), round(y))
    matrix.translate(x, y)

    if mirrored:
        m = list(matrix)
        m[0] = -m[0]
        m[2] = -m[2]
        matrix = cairo.Matrix(*m)

    return matrix


class CanvasRenderer (Gtk.DrawingArea, DrawCursorMixin):
    """Render the document model to screen.

    Can render the document in a transformed way, including translation,
    scaling and rotation.

    """

    ## Method defs

    def __init__(self, tdw, idle_redraw_priority=None):
        super(CanvasRenderer, self).__init__()
        self.init_draw_cursor()

        self.connect("draw", self._draw_cb)
        self._idle_redraw_priority = idle_redraw_priority
        self._idle_redraw_queue = []
        self._idle_redraw_src_id = None

        self.connect("state-changed", self._state_changed_cb)

        self._tdw = tdw

        # Currently we don't need a window while we're packed into the
        # TiledDrawWidget (a GtkEventBox subclass).
        # Capturing events via a separate EventBox separates concerns
        # and theoretically makes this the renderer component pluggable.
        # GtkDrawingArea's PyGI implementation still uses no_window,
        # unlike other GI flavours, and unlike C-style gtk. We want the
        # standard behaviour though.
        self.set_has_window(False)

        self.visualize_rendering = False

        self.translation_x = 0.0
        self.translation_y = 0.0
        self.scale = 1.0
        self.rotation = 0.0
        self.mirrored = False
        self.cached_transformation_matrix = None
        self.display_filter = None

        self.overlay_layer = None

        # gets overwritten for the main window
        self.zoom_max = 5.0
        self.zoom_min = 1 / 5.0

        # Sensitivity; we draw via a cached snapshot while the widget is
        # insensitive. tdws are generally only insensitive during loading and
        # saving, and because we now process the GTK main loop during loading
        # and saving, we need to avoid drawing partially-loaded files.
        self._insensitive_state_content = None

        # Overlays
        self.model_overlays = []
        self.display_overlays = []

        # Pixelize at high zoom-ins.
        # The icon editor needs to be able to adjust this.
        self.pixelize_threshold = 2.8

        # Backgroundless rendering
        self._real_alpha_check_pattern = None
        self._fake_alpha_check_tile = None
        self._init_alpha_checks()

        # Higher-quality mipmap choice
        # Turn off for a speedup during dragging or scrolling
        self._hq_rendering = True
        self._restore_hq_rendering_timeout_id = None

        self.connect("configure-event", self._configure_event_cb)

    def _init_alpha_checks(self):
        """Initialize the alpha check backgrounds"""
        # Real: checkerboard pattern, rendered via Cairo
        assert tiledsurface.N % gui.style.ALPHA_CHECK_SIZE == 0
        n = tiledsurface.N
        size = gui.style.ALPHA_CHECK_SIZE
        nchecks = int(n // size)
        cairo_surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, n, n)
        cr = cairo.Context(cairo_surf)
        render_checks(cr, size, nchecks)
        cairo_surf.flush()
        # Real: MyPaint background surface for layers-but-no-bg rendering
        pattern = cairo.SurfacePattern(cairo_surf)
        pattern.set_extend(cairo.EXTEND_REPEAT)
        self._real_alpha_check_pattern = pattern
        # Fake: faster rendering, but ugly
        tile = np.empty((n, n, 4), dtype='uint16')
        f = 1 << 15
        col1 = [int(f * c) for c in gui.style.ALPHA_CHECK_COLOR_1] + [f]
        col2 = [int(f * c) for c in gui.style.ALPHA_CHECK_COLOR_2] + [f]
        tile[:] = col1
        for i in xrange(nchecks):
            for j in xrange(nchecks):
                if (i + j) % 2 == 0:
                    continue
                ia, ib = (i * size), ((i + 1) * size)
                ja, jb = (j * size), ((j + 1) * size)
                tile[ia:ib, ja:jb] = col2
        self._fake_alpha_check_tile = tile

    @property
    def app(self):
        return self._tdw.app

    @property
    def doc(self):
        return self._tdw.doc

    @event
    def transformation_updated(self):
        """Event: transformation was updated"""

    def _invalidate_cached_transform_matrix(self):
        self.cached_transformation_matrix = None

    def _get_x(self):
        return self._translation_x

    def _set_x(self, val):
        self._translation_x = val
        self._invalidate_cached_transform_matrix()
    translation_x = property(_get_x, _set_x)

    def _get_y(self):
        return self._translation_y

    def _set_y(self, val):
        self._translation_y = val
        self._invalidate_cached_transform_matrix()
    translation_y = property(_get_y, _set_y)

    def _get_scale(self):
        return self._scale

    def _set_scale(self, val):
        self._scale = val
        self._invalidate_cached_transform_matrix()
    scale = property(_get_scale, _set_scale)

    def _get_rotation(self):
        return self._rotation

    def _set_rotation(self, val):
        self._rotation = val
        self._invalidate_cached_transform_matrix()
    rotation = property(_get_rotation, _set_rotation)

    def _get_mirrored(self):
        return self._mirrored

    def _set_mirrored(self, val):
        self._mirrored = val
        self._invalidate_cached_transform_matrix()
    mirrored = property(_get_mirrored, _set_mirrored)

    def _state_changed_cb(self, widget, oldstate):
        """Handle the sensitivity state changing

        Saving and loading images toggles the sensitivity state on all
        toplevel windows. This causes a state shift on the TDW too.
        While the TDW is insensitive, its cursor is updated to respect
        the toplevel's cursor (typically a watch or an hourglass or
        something).

        """
        insensitive = self.get_state_flags() & Gtk.StateFlags.INSENSITIVE
        if insensitive and (not self._insensitive_state_content):
            alloc = widget.get_allocation()
            w = alloc.width
            h = alloc.height
            surface = self._new_image_surface_from_visible_area(0, 0, w, h)
            self._insensitive_state_content = surface
        elif (not insensitive) and self._insensitive_state_content:
            self._insensitive_state_content = None
        self.update_cursor()

    ## Redrawing

    def canvas_modified_cb(self, model, x, y, w, h):
        """Handles area redraw notifications from the underlying model"""

        if self._insensitive_state_content:
            return False

        if not self.get_window():
            return

        if w == 0 and h == 0:
            # Full redraw (used when background has changed).
            # logger.debug('Full redraw')
            self.queue_draw()
            return

        # Create an expose event with the event bbox rotated/zoomed.
        corners = [(x, y), (x + w, y), (x, y + h), (x + w, y + h)]
        corners = [self.model_to_display(x, y) for (x, y) in corners]
        bbox = helpers.rotated_rectangle_bbox(corners)
        self.queue_draw_area(*bbox)

    def queue_draw(self):
        if self._idle_redraw_priority is None:
            super(CanvasRenderer, self).queue_draw()
            return
        self._queue_idle_redraw(None)

    def queue_draw_area(self, x, y, w, h):
        if self._idle_redraw_priority is None:
            super(CanvasRenderer, self).queue_draw_area(x, y, w, h)
            return
        bbox = helpers.Rect(x, y, w, h)
        self._queue_idle_redraw(bbox)

    def _queue_idle_redraw(self, bbox):
        queue = self._idle_redraw_queue
        if bbox is None:
            queue[:] = []
        elif None in queue:
            return
        else:
            queue[:] = [b for b in queue if not bbox.contains(b)]
            for b in queue:
                if b.contains(bbox):
                    return
        queue.append(bbox)
        if self._idle_redraw_src_id is not None:
            return
        src_id = GLib.idle_add(
            self._idle_redraw_cb,
            priority = self._idle_redraw_priority,
        )
        self._idle_redraw_src_id = src_id

    def _idle_redraw_cb(self):
        assert self._idle_redraw_src_id is not None
        queue = self._idle_redraw_queue
        if len(queue) > 0:
            bbox = queue.pop(0)
            if bbox is None:
                super(CanvasRenderer, self).queue_draw()
            else:
                super(CanvasRenderer, self).queue_draw_area(*bbox)
        if len(queue) == 0:
            self._idle_redraw_src_id = None
            return False
        return True

    ## Redraw events

    def current_layer_changed_cb(self, rootstack, path):
        self.update_cursor()

    def layer_props_changed_cb(self, rootstack, path, layer, changed):
        self.update_cursor()

    def frame_enabled_changed_cb(self, model, enabled):
        self.queue_draw()

    def frame_updated_cb(self, model, old_frame, new_frame):
        self.queue_draw()

    ## Transformations and coords

    def display_to_model(self, disp_x, disp_y):
        """Converts display coordinates to model coordinates.
        """
        matrix = cairo.Matrix(*self._get_model_view_transformation())
        assert not matrix.invert()
        view_model = matrix
        return view_model.transform_point(disp_x, disp_y)

    def model_to_display(self, model_x, model_y):
        """Converts model coordinates to display coordinates.
        """
        model_view = self._get_model_view_transformation()
        return model_view.transform_point(model_x, model_y)

    def _get_model_view_transformation(self):
        if self.cached_transformation_matrix is None:
            scale_factor = self.get_scale_factor()
            # HiDPI: logical "device" (widget) pixels to screen pixels
            matrix = calculate_transformation_matrix(
                self.scale / scale_factor, self.rotation,
                self.translation_x, self.translation_y,
                self.mirrored
            )
            self.cached_transformation_matrix = matrix
            self.transformation_updated()
        return self.cached_transformation_matrix

    def _configure_event_cb(self, widget, event):
        # The docs say to handle this on the toplevel to be notified of
        # HiDPI scale_factor changes. Not sure yet whether this will
        # work down at this level - I've no fancy HiDPI hardware to test
        # with. For now, just try this.
        logger.debug("configure-event received. Invalidating transform")
        self._invalidate_cached_transform_matrix()
        self.queue_draw()

    def is_translation_only(self):
        return self.rotation == 0.0 and self.scale == 1.0 and not self.mirrored

    def pick_color(self, x, y, size=3):
        """Picks the rendered colour at a particular point.

        :param int x: X coord of pixel to pick (widget/device coords)
        :param int y: Y coord of pixel to pick (widget/device coords)
        :param int size: Size of the sampling square.
        :returns: The colour sampled.
        :rtype: lib.color.UIColor

        This method operates by rendering part of the document using the
        current settings, then averaging the colour values of the pixels
        within the sampling square.

        """
        # TODO: the ability to turn *off* this kind of "sample merged".
        # Ref: https://github.com/mypaint/mypaint/issues/333

        # Make a square surface for the sample.
        size = max(1, int(size))

        # Extract a square Cairo surface containing the area to sample
        r = int(size // 2)
        x = int(x) - r
        y = int(y) - r
        surf = self._new_image_surface_from_visible_area(
            x, y,
            size, size,
            use_filter = False,
        )

        # Extract a pixbuf, then an average color.
        # surf.write_to_png("/tmp/grab.png")
        pixbuf = Gdk.pixbuf_get_from_surface(surf, 0, 0, size, size)
        color = lib.color.UIColor.new_from_pixbuf_average(pixbuf)
        return color

    def _new_image_surface_from_visible_area(self, x, y, w, h,
                                             use_filter=True):
        """Render part of the doc to a new cairo image surface, as seen.

        :param int x: Rectangle left edge (widget/device coords)
        :param int y: Rectangle top edge (widget/device coords)
        :param int w: Rectangle width (widget/device coords)
        :param int h: Rectangle height (widget/device coords)
        :param bool use_filter: Apply display filters to rendering.
        :rtype: cairo.ImageSurface
        :returns: A rendered of the document, as seen on screen

        Creates and returns a new cairo.ImageSurface of the given size,
        containing an image of the document as it would appears on
        screen within the given rectangle. The area to extract is a
        rectangle in display (widget) coordinates, but it doesn't
        actually have to be within the visible area. Used for
        snapshotting and sampling colours.

        """

        # Start with a clean black slate.
        x = int(x)
        y = int(y)
        w = max(1, int(w))
        h = max(1, int(h))
        surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
        cr = cairo.Context(surf)
        cr.set_source_rgb(0, 0, 0)
        cr.paint()

        # The rendering routines used are those used by the normal draw
        # handler, so we need an offset before calling them.
        cr.translate(-x, -y)

        # Paint checkerboard if we won't be rendering an opaque background
        model = self.doc
        render_is_opaque = model and model.layer_stack.get_render_is_opaque()
        if (not render_is_opaque) and self._draw_real_alpha_checks:
            cr.set_source(self._real_alpha_check_pattern)
            cr.paint()
        if not model:
            return surf

        # Render just what we need.
        transformation, surface, sparse, mipmap_level, clip_rect = \
            self._render_prepare(cr)
        display_filter = None
        if use_filter:
            display_filter = self.display_filter
        self._render_execute(
            cr,
            transformation,
            surface,
            sparse,
            mipmap_level,
            clip_rect,
            filter = display_filter,
        )
        surf.flush()
        return surf

    @property
    def _draw_real_alpha_checks(self):
        if not self.app:
            return True
        return self.app.preferences["view.real_alpha_checks"]

    def _draw_cb(self, widget, cr):
        """Draw handler"""

        # Don't render any partial views of the document if the widget
        # isn't sensitive to user input. If we don't do this, loading a
        # big doc might show each layer individually as it loads.
        if self._insensitive_state_content:
            cr.set_source_surface(self._insensitive_state_content, 0, 0)
            cr.paint()
            return True

        # Paint checkerboard if we won't be rendering an opaque background
        model = self.doc
        render_is_opaque = model and model.layer_stack.get_render_is_opaque()
        if (not render_is_opaque) and self._draw_real_alpha_checks:
            cr.set_source(self._real_alpha_check_pattern)
            cr.paint()
        if not model:
            return True

        # Paint a random grey behind what we're about to render
        # if visualization is needed.
        if self.visualize_rendering:
            tmp = random.random()
            cr.set_source_rgb(tmp, tmp, tmp)
            cr.paint()

        # Prep a pixbuf-surface aligned to the model to render into.
        # This also applies the transformation.
        transformation, surface, sparse, mipmap_level, clip_rect = \
            self._render_prepare(cr)

        # not sure if it is a good idea to clip so tightly
        # has no effect right now because device_bbox is always smaller
        model_bbox = surface.x, surface.y, surface.w, surface.h
        cr.rectangle(*model_bbox)
        cr.clip()

        # Clear the pixbuf to be rendered with a random red,
        # to make it apparent if something is not being painted.
        if self.visualize_rendering:
            surface.pixbuf.fill(int(random.random() * 0xff) << 16)

        # Render to the pixbuf, then paint it.
        self._render_execute(
            cr,
            transformation,
            surface,
            sparse,
            mipmap_level,
            clip_rect,
            filter = self.display_filter,
        )

        # Using different random blues helps make one rendered bbox
        # distinct from the next when the user is painting.
        if self.visualize_rendering:
            cr.set_source_rgba(0, 0, random.random(), 0.4)
            cr.paint()

        # Model coordinate space:
        cr.restore()  # CONTEXT2<<<
        for overlay in self.model_overlays:
            cr.save()
            overlay.paint(cr)
            cr.restore()

        # Back to device coordinate space
        cr.restore()  # CONTEXT1<<<
        for overlay in self.display_overlays:
            cr.save()
            overlay.paint(cr)
            cr.restore()

        return True

    def _render_get_clip_region(self, cr, device_bbox):
        """Get the area that needs to be updated, in device coords.

        Called when handling "draw" events.  This uses Cairo's clip
        region, which ultimately derives from the areas sent by
        lib.document.Document.canvas_area_modified().  These can be
        small or large if the update comes form the user drawing
        something. When panning or zooming the canvas, it's a full
        redraw, and the area corresponds to the entire display.

        :param cairo.Context cr: as passed to the "draw" event handler.
        :param tuple device_bbox: (x,y,w,h) widget extents
        :returns: (clipregion, sparse)
        :rtype: tuple

        The clip region return value is a lib.helpers.Rect containing
        the area to redraw in display coordinates, or None.

        This also determines whether the redraw is "sparse", meaning
        that the clip region returned does not contain the centre of the
        device bbox.

        """

        # Could this be an alternative?
        # x0, y0, x1, y1 = cr.clip_extents()
        # sparse = True
        # clip_region = (x0, y0, x1-x0, y1-y0)
        # return clip_region, sparse

        # Get the area which needs to be updated, in device coordinates, and
        # determine whether the render is "sparse", [TODO: define what this
        # means]
        x, y, w, h = device_bbox
        cx, cy = x + w // 2, y + h // 2

        # As of 2012-07-08, Ubuntu Precise (LTS, unfortunately) and Debian
        # unstable(!) use python-cairo 1.8.8, which does not support
        # the cairo.Region return from Gdk.Window.get_clip_region() we
        # really need. They'll be important to support for a while, so we
        # have to use an inefficient workaround using the complete clip
        # rectangle for the update.
        #
        # http://stackoverflow.com/questions/6133622/
        # http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=653588
        # http://packages.ubuntu.com/python-cairo
        # http://packages.debian.org/python-cairo
        #
        # Update 2015-08-14: actually, on Debian-derived systems you
        # need gir1.2-freedesktop 1.44.0, which contains a wrapping for
        # some cairoish return values from GTK/GDK, including
        # CairoRegion: cairo-1.0.typelib. On MSYS2, ensure you have
        # mingw-w64-i686-gobject-introspection-runtime or its x86_64
        # equivalent installed, also at version 1.44.

        clip_exists, rect = Gdk.cairo_get_clip_rectangle(cr)
        if clip_exists:
            # It's a wrapped cairo_rectangle_int_t, CairoRectangleInt
            # Convert to a better representation for our purposes,
            # noting https://github.com/mypaint/mypaint/issues/433
            rect = helpers.Rect(rect.x, rect.y, rect.width, rect.height)
            sparse = (
                cx < rect.x
                or cx > (rect.x + rect.w)
                or cy < rect.y
                or cy > (rect.y + rect.h)
            )
        else:
            rect = None
            sparse = False

        return rect, sparse

    def _tile_is_visible(self, tx, ty, transformation, clip_rect,
                         translation_only):
        """Tests whether an individual tile is visible.

        This is sometimes worth doing during rendering, but not always.
        Currently we use _render_get_clip_region()'s sparse flag to
        determine whether this is necessary. The original logic for
        this function was documented as...

        > it is worth checking whether this tile really will be visible
        > (to speed up the L-shaped expose event during scrolling)
        > (speedup clearly visible; slowdown measurable when always
        > executing this code)

        I'm not 100% certain that GTK3 does panning redraws this way,
        so perhaps this method is uneccessary for those?
        However this method is always used when rendering during
        painting, or other activities that send partial updates.

        """
        n = tiledsurface.N
        if translation_only:
            x, y = transformation.transform_point(tx * n, ty * n)
            bbox = (int(x), int(y), n, n)
        else:
            corners = [
                (tx * n, ty * n),
                ((tx + 1) * n, ty * n),
                (tx * n, (ty + 1) * n),
                ((tx + 1) * n, (ty + 1) * n),
            ]
            corners = [
                transformation.transform_point(x_, y_)
                for (x_, y_) in corners
            ]
            bbox = helpers.rotated_rectangle_bbox(corners)
        tile_rect = helpers.Rect(*bbox)
        return clip_rect.overlaps(tile_rect)

    def _render_prepare(self, cr):
        """Prepares a blank pixbuf & other details for later rendering.

        Called when handling "draw" events. The size and shape of the
        returned pixbuf (wrapped in a tile-accessible and read/write
        lib.pixbufsurface.Surface) is determined by the Cairo clipping
        region that expresses what we've been asked to redraw, and by
        the TDW's own view transformation of the document.

        """
        # Determine what to draw, and the nature of the reveal.
        allocation = self.get_allocation()
        w, h = allocation.width, allocation.height
        device_bbox = (0, 0, w, h)
        clip_rect, sparse = self._render_get_clip_region(cr, device_bbox)
        x, y, w, h = device_bbox

        # Use a copy of the cached translation matrix for this
        # rendering. It'll need scaling if using a mipmap level
        # greater than zero.
        transformation = cairo.Matrix(*self._get_model_view_transformation())

        # HQ rendering causes a very clear slowdown on some hardware.
        # Probably could avoid this entirely by rendering differently,
        # but for now, if the canvas is being panned around,
        # just render more simply.
        if self._hq_rendering:
            mipmap_level = max(0, int(floor(log(1 / self.scale, 2))))
        else:
            mipmap_level = max(0, int(ceil(log(1 / self.scale, 2))))

        # OPTIMIZE: If we would render tile scanlines,
        # OPTIMIZE:  we could probably use the better one above...
        mipmap_level = min(mipmap_level, tiledsurface.MAX_MIPMAP_LEVEL)
        transformation.scale(2**mipmap_level, 2**mipmap_level)

        # bye bye device coordinates
        cr.save()   # >>>CONTEXT1
        cr.transform(transformation)
        cr.save()   # >>>CONTEXT2

        # calculate the final model bbox with all the clipping above
        x1, y1, x2, y2 = cr.clip_extents()
        if not self.is_translation_only():
            # Looks like cairo needs one extra pixel rendered for
            # interpolation at the border. If we don't do this, we get dark
            # stripe artifacts when panning while zoomed.
            x1 -= 1
            y1 -= 1
            x2 += 1
            y2 += 1
        x1, y1 = int(floor(x1)), int(floor(y1))
        x2, y2 = int(ceil(x2)), int(ceil(y2))

        # We always render with alpha to get hardware acceleration,
        # even when we could avoid using the alpha channel. Speedup
        # factor 3 for ATI/Radeon Xorg driver (and hopefully others).
        # https://bugs.freedesktop.org/show_bug.cgi?id=28670

        surface = pixbufsurface.Surface(x1, y1, x2 - x1 + 1, y2 - y1 + 1)
        return transformation, surface, sparse, mipmap_level, clip_rect

    def _render_execute(self, cr, transformation, surface, sparse,
                        mipmap_level, clip_rect, filter=None):
        """Renders tiles into a prepared pixbufsurface, then blits it.


        """
        translation_only = self.is_translation_only()

        if self.visualize_rendering:
            surface.pixbuf.fill(int(random.random() * 0xff) << 16)

        fake_alpha_check_tile = None
        if not self._draw_real_alpha_checks:
            fake_alpha_check_tile = self._fake_alpha_check_tile

        # Determine which tiles to render.
        tiles = list(surface.get_tiles())
        if sparse:
            tiles = [
                (tx, ty) for (tx, ty) in tiles
                if self._tile_is_visible(
                    tx,
                    ty,
                    transformation,
                    clip_rect,
                    translation_only,
                )
            ]

        # Composite each stack of tiles in the exposed area
        # into the pixbufsurface.
        self.doc._layers.render(
            surface,
            tiles,
            mipmap_level,
            overlay = self.overlay_layer,
            opaque_base_tile = fake_alpha_check_tile,
            filter = filter,
        )

        # Set the surface's underlying pixbuf as the source, then paint
        # it with Cairo. We don't care if it's pixelized at high zoom-in
        # levels: in fact, it'll look sharper and better.
        Gdk.cairo_set_source_pixbuf(
            cr, surface.pixbuf,
            round(surface.x), round(surface.y)
        )
        if self.scale > self.pixelize_threshold:
            pattern = cr.get_source()
            pattern.set_filter(cairo.FILTER_NEAREST)
        cr.paint()

    def scroll(self, dx, dy, ongoing=True):
        self.translation_x -= dx
        self.translation_y -= dy
        if ongoing:
            self.defer_hq_rendering()
        self.queue_draw()

        # This speeds things up nicely when scrolling is already
        # fast, but produces temporary artifacts and an
        # annoyingly non-constant framerate otherwise.
        #
        # self.window.scroll(int(-dx), int(-dy))
        #
        # It might be worth it if it was done only once per
        # redraw, instead of once per motion event. Maybe try to
        # implement something like "queue_scroll" with priority
        # similar to redraw? (The GTK commit responsible for bug
        # http://bugzilla.gnome.org/show_bug.cgi?id=702392 might
        # solve this problem, I think.)

    def get_center(self):
        """Return the center position in display coordinates.
        """
        alloc = self.get_allocation()
        return (alloc.width / 2.0, alloc.height / 2.0)

    def get_center_model_coords(self):
        """Return the center position in model coordinates.
        """
        center = self.get_center()
        return self.display_to_model(*center)

    def recenter_document(self):
        """Recentres the view onto the document's centre.
        """
        x, y, w, h = self.doc.get_effective_bbox()
        cx = x + (w / 2.0)
        cy = y + (h / 2.0)
        self.recenter_on_model_coords(cx, cy)

    def recenter_on_model_coords(self, cx, cy):
        """Recentres the view to a specified point, in model coordinates.
        """
        dcx, dcy = self.model_to_display(cx, cy)
        self.recenter_on_display_coords(dcx, dcy)

    def recenter_on_display_coords(self, cx, cy):
        current_cx, current_cy = self.get_center()
        self.translation_x += current_cx - cx
        self.translation_y += current_cy - cy
        self.queue_draw()

    def defer_hq_rendering(self, t=1.0 / 8):
        """Use faster but lower-quality rendering for a brief period

        :param float t: The time to defer for, in seconds

        This method is intended to be called repeatedly
        from scroll or drag event handlers,
        or other times when the entire display
        may need to be redrawn repeatedly in short order.
        It turns off normal rendering, and updates the future time
        at which normal rendering will be automatically resumed.
        Resumption of normal service entails a full redraw,
        so choose `t` appropriately.

        Normal rendering looks better (it uses a better mipmap),
        and it's OK for most screen updates.
        However it's slow enough to make rendering
        lag appreciably when scrolling.

        """
        if self._restore_hq_rendering_timeout_id:
            GLib.source_remove(self._restore_hq_rendering_timeout_id)
            self._restore_hq_rendering_timeout_id = None
        else:
            logger.debug("hq_rendering: deferring for %0.3fs...", t)
            self._hq_rendering = False
        self._restore_hq_rendering_timeout_id = GLib.timeout_add(
            interval = int(t * 1000),
            function = self._resume_hq_rendering_timeout_cb,
        )

    def _resume_hq_rendering_timeout_cb(self):
        self._hq_rendering = True
        self.queue_draw()
        self._restore_hq_rendering_timeout_id = None
        logger.debug("hq_rendering: resumed")
        return False


## Testing


def _make_testbed_model():
    import lib.brush
    import lib.document
    brush = lib.brush.BrushInfo()
    brush.load_defaults()
    return lib.document.Document(brush)


def _test():
    from document import CanvasController
    from freehand import FreehandMode
    model = _make_testbed_model()
    tdw = TiledDrawWidget()
    tdw.set_model(model)
    tdw.set_size_request(640, 480)
    tdw.renderer.visualize_rendering = True
    ctrlr = CanvasController(tdw)
    ctrlr.init_pointer_events()
    ctrlr.modes.default_mode_class = FreehandMode
    win = Gtk.Window()
    win.set_title("tdw test")
    win.connect("destroy", lambda *a: Gtk.main_quit())
    win.add(tdw)
    win.show_all()
    Gtk.main()
    model.cleanup()


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    _test()



# This file is part of MyPaint.
# Copyright (C) 2011-2015 by Andrew Chadwick <a.t.chadwick@gmail.com>
# Copyright (C) 2007-2012 by Martin Renold <martinxyz@gmx.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import division, print_function


def make_test_stack():
    """Makes a simple test RootLayerStack (2 branches of 3 leaves each)

    :return: The root stack, and a list of its leaves.
    :rtype: tuple

    """
    import lib.layer.group
    import lib.layer.data
    import lib.layer.tree
    root = lib.layer.tree.RootLayerStack(doc=None)
    layer0 = lib.layer.group.LayerStack(name='0')
    root.append(layer0)
    layer00 = lib.layer.data.PaintingLayer(name='00')
    layer0.append(layer00)
    layer01 = lib.layer.data.PaintingLayer(name='01')
    layer0.append(layer01)
    layer02 = lib.layer.data.PaintingLayer(name='02')
    layer0.append(layer02)
    layer1 = lib.layer.group.LayerStack(name='1')
    root.append(layer1)
    layer10 = lib.layer.data.PaintingLayer(name='10')
    layer1.append(layer10)
    layer11 = lib.layer.data.PaintingLayer(name='11')
    layer1.append(layer11)
    layer12 = lib.layer.data.PaintingLayer(name='12')
    layer1.append(layer12)
    return (root, [layer00, layer01, layer02, layer10, layer11, layer12])



# This file is part of MyPaint.
# Copyright (C) 2015 by Andrew Chadwick <a.t.chadwick@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Error classes which may be raised by gui-independent code"""

from __future__ import division, print_function


class FileHandlingError (Exception):
    """Simple problem loading or saving files; user-facing string

    Covers expected things like missing required path elements, missing
    files, unsupported formats etc.

    The stringification of a FileHandlingError should always be
    presentable to the user directly, and marked for translation at the
    point where it's raised. It may contain diagnostic information, but
    should always be prefixed or suffixed with gloss for ordinary users.

    In general, if one of these is raised as a response to another
    exception, log that error with (yourmodule.logger.exception()) with
    programmer-focussed diagnostic info, and favour user-presentable
    info for the message string.

    """

    def __init__(self, msg, investigate_dir=None):
        super(FileHandlingError, self).__init__(msg)
        self.investigate_dir = investigate_dir


class AllocationError (Exception):
    """Indicates a failure to construct a required internal object.

    This may be used as a stopgap to cover probable out-of-memory
    conditions which GI hasn't wrapped as nice Pythonic `MemoryError`s.
    The GdkPixbuf wrappers are quite prone to this.

    The stringification of an AllocationError should always be
    presentable to the user directly, and marked for translation at the
    point where it's raised. It may contain diagnostic information, but
    should always be prefixed or suffixed with gloss for ordinary users.

    In general, if one of these is raised as a response to another
    exception, log that error with (yourmodule.logger.exception()) with
    programmer-focussed diagnostic info, and favour user-presentable
    info for the message string.

    """






# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2017 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Mixin and decorator functionality for the MVP UI pattern."""

# Imports:

import os
import inspect
import abc
import functools
import logging

from lib.gibindings import Gtk


logger = logging.getLogger(__name__)


# Main API classes:

class Presenter:
    """Abstract interface for standardized MVP presenters.

    What's MVP?
    -----------

    "Model-View-Presenter" or "MVP" is a UI pattern where a presenter
    object mediates between a View and a Model. The presenter observes
    changes to both sides, and for each change, updates the other side.
    Presenters encapsulate presentation logic and formatting for a set
    of displayable objects, termed the "view". They also turn user
    actions into internal state changes of obejcts which cannot by
    themselves be shown to the user, called the "model".

    Ref: https://en.wikipedia.org/wiki/Model-view-presenter

    MyPaint MVP conventions
    -----------------------

    In MyPaint, a Presenter's View is a hierarchy of basic GTK objects
    that are typically laid out with the Glade UI designer and stored as
    an XML file. View observation is done with GTK signal callback
    methods, which are named in the XML UI definition, and which are
    expected to be implemented by the corresponding Presenter object.

    In MyPaint, a Presenter's Model can be anything. It's normally one
    or more objects from lib, for example a layer or a whole layers
    stack.  Observation of model objects happens by attaching callbacks
    to their "@event" methods (see lib.observable).

    In the suggested implementation, presenters are tightly bound to
    their view, and often own and instantiate it on demand. See
    BuiltUIPresenter for a mixin class that makes this easy.  They may
    be quite loosely bound to their model if needed, and are typically
    introduced to it during construction. The rest of the code,
    including other presenters, should keep a ref to the presenter for
    as long as the presentation logic needs to happen.

    It is conventional here to name concrete presenter classes like
    "<Model>UI" or "<Role>UI", noting that a presenter provides the
    behavioural aspects of the user interface.

    MyPaint encourages tight-ish coupling of a View to its Presenter(s)
    but looser coupling of Presenters to their Model objects. Presenters
    can own their View. Conversely, presenters almost never own very
    complicated models: most use their view to present aspects of a
    model owned by something else.

    For a presenter observing its model hierarchy, connect methods to
    "@event"s exposed by the highest level model object you can find.
    See "lib.observable" for how this makes garbage collection nicer.
    Conversely, when a presenter needs to observe its view herarchy, use
    standard GTK signals and individual widgets' connect() method.

    Decorators
    ----------

    Use the @model_updater and @view_updater decorators on the
    Presenter's callback methods to save on having to write tons of
    fiddly value tests or implementing other ways of preventing loops.

    Decorate each model observer callback and each view signal handler
    callback to make sure you don't end up with a cascade of calls.
    The callbacks can be specified with or without args.

    >>> class SomeUI (Presenter):
    ...
    ...     @view_updater
    ...     def model_field_A_updated_cb(self, *args, **kwargs):
    ...         pass
    ...
    ...     @view_updater(default=42)
    ...     def model_field_B_updated_cb(self, *args, **kwargs):
    ...         pass
    ...
    ...     @view_updater
    ...     def view_widget_1_updated_cb(self, *args, **kwargs):
    ...         pass
    ...
    ...     @model_updater(default=False)
    ...     def view_widget_2_updated_cb(self, *args, **kwargs):
    ...         pass

    Why do this? It follows from the MVP pattern that if you're not
    careful about the update flow, you can end up with a loop.

    These decorators fix that potential problem by completely skipping
    the wrapped methods if the Presenter is currently doing the *other*
    type of update.  You can return a default value you specify if
    needed: this can keep GTK+ signals happier.

    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def widget(self):
        """The primary widget of the view.

        :returns: the main view widget
        :rtype: Gtk.Widget

        All presenters must expose a single widget to be the primary
        entry point to the view. This property should be overridden to
        provide, for example, the widget to use for the contents of a
        fancy dialog, or the toplevel window in an application.

        For "adoptive" presenters (ones where the Presenter adopts a
        given widget hierarchy as its view, and connects signals to it),
        make the "widget" property *read-write* and do the extra hookup
        work in your "@widget.setter". Your presenter's owner then has a
        clean interface for attaching bundles of additional behaviour to
        parts of its own view.

        """

    @staticmethod
    def _updater(func=None, default=None):
        """Decorates methods that must only be called once at any one time.

        :param callable func: The method to be wrapped.
        :param default: Return value for when the the method is skipped.
        :returns: the decorated method.

        The wrapped function is skipped if it is currently being called
        on the presenter.

        Use this decorator (as model_updater or view_updater) for all
        model or view observer callbacks which update the other side.
        Doing this prevents potential loops.

        See also: lib.observable.event, Gtk.Widget.connect.

        """

        # Allow the @pie syntax to be used with or without args.
        if func is None:
            return functools.partial(Presenter._updater, default=default)

        @functools.wraps(func)
        def method_wrapper(inst, *args, **kwargs):
            try:
                in_call = func.__in_call
            except AttributeError:
                in_call = False

            if in_call:
                logger.debug("@_updater: suppressed a call to %r", func)
                return default

            func.__in_call = True
            try:
                return func(inst, *args, **kwargs)
            except:
                raise
            finally:
                func.__in_call = False

        return method_wrapper


class BuiltUIPresenter (Presenter):
    """Mixin providing Pythonic access to views built from GtkBuilder XML.

    This style of presenter has its view defined entirely by a
    corresponding GTK+ UI XML file. The view is constructed on demand by
    methods here, and its signals will be automatically bound to methods
    of the Presenter when the view objects are constructed. The
    Presenter mixin interface provides overridable hooks for setting the
    initial state of the view.

    Subclasses should define `primary_widget` so as to access and return
    a widget object from `view`.

    """

    _UI_FILE_EXTENSIONS = [".ui", ".glade", ".xml"]

    @property
    def view(self):
        """On-demand access the built View objects by attribute lookup.

        :returns: A wrapper for accessing UI objects.
        :rtype: _ViewWrapper

        When accessed for the first time, this property method
        constructs an internal GtkBuilder, and uses it to instantiate
        GTK objects from a UI XML file.

        The file is expected to reside in the same directory as
        self.__class__.__file___, and to be named after it (*.glade).
        However, if self.__ui_xml__ is defined, that will be used for
        the basename instead.

        Upon construction, the builder connects the new objects' signals
        to self.

        """

        # Cache the loaded UI wrapper.
        try:
            return self.__ui_wrapper
        except AttributeError:
            pass

        # Select a .glade file
        class_defn_file = inspect.getfile(self.__class__)
        mod_dirname = os.path.dirname(class_defn_file)
        try:
            ui_file_basenames = [self.__ui_xml__]
        except AttributeError:
            py_basename = os.path.basename(class_defn_file)
            py_basename, _oldext = os.path.splitext(py_basename)
            ui_file_basenames = [
                py_basename + e
                for e in self._UI_FILE_EXTENSIONS
            ]

        # Load objects, bind, cache and return.
        for basename in ui_file_basenames:
            ui_path = os.path.join(mod_dirname, basename)
            logger.debug("BuiltUIPresenter: trying to load %r", ui_path)
            if not os.path.isfile(ui_path):
                continue
            logger.debug(
                "BuiltUIPresenter: found UI definition in %r; loading",
                ui_path,
            )
            wrapper = _ViewWrapper(ui_path)
            self.__ui_wrapper = wrapper
            self.init_view()
            wrapper._builder.connect_signals(self)
            self.init_view_post_connect()
            return wrapper

        raise RuntimeError(
            "BuiltUIPresenter: could not load "
            "any UI definition XML file (tried %r, in %r)"
            % (ui_file_basenames, mod_dirname),
        )

    def init_view(self):
        """Hook: initialize the view objects (before signal connection).

        Implementations should set the initial state of all relevant UI
        object to reflect the state of the model. It is called before
        signals are connected, for convenience.

        Note that this method is called the first time the view()
        property is accessed.

        This base implementation does nothing. You'll typically want to
        use this rather than init_view_post_connect().

        """

    def init_view_post_connect(self):
        """Hook: initialize the view objects (AFTER signal connection).

        This method is called after connection of signals, but is
        otherwise identical to init_view(). Override that method
        instead, unless you have specific needs.

        """


# Convenience/selfdoc aliases for the updater decorator:

model_updater = Presenter._updater
view_updater = Presenter._updater


# Helper classes:

class _ViewWrapper:
    """Private GTK+ view object abstraction.

    This can be accessed by the ID of object IDs inside the
    corresponding UI XML file.

    """

    _TRANSLATION_DOMAIN = "mypaint"

    def __init__(self, filename):
        self._filename = filename
        builder = Gtk.Builder()
        builder.set_translation_domain(self._TRANSLATION_DOMAIN)
        builder.add_from_file(filename)
        self._builder = builder
        self._cache = {}

    def __getattr__(self, attr_name):
        try:
            return self._cache[attr_name]
        except KeyError:
            pass

        for name in (attr_name, attr_name.replace("_", "-")):
            obj = self._builder.get_object(name)
            if obj is not None:
                self._cache[attr_name] = obj
                return obj

        raise AttributeError(
            "No object with name %r (incl. \"_\" substs) in %r"
            % (attr_name, self._filename),
        )



# This file is part of MyPaint.
# Copyright (C) 2011-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""The application toolbar, and its specialised widgets"""


## Imports

from __future__ import division, print_function
import os
from gettext import gettext as _

from lib.gibindings import Gtk

from . import widgets


## Module constants


FRAMEWORK_XML = 'toolbar.xml'
MERGEABLE_XML = [
    ("toolbar1_file", 'toolbar-file.xml', _("File handling")),
    ("toolbar1_scrap", 'toolbar-scrap.xml', _("Scraps switcher")),
    ("toolbar1_edit", 'toolbar-edit.xml', _("Undo and Redo")),
    ("toolbar1_blendmodes", 'toolbar-blendmodes.xml', _("Blend Modes")),
    ("toolbar1_linemodes", 'toolbar-linemodes.xml', _("Line Modes")),
    ("toolbar1_view_modes", 'toolbar-view-modes.xml', _("View (Main)")),
    ("toolbar1_view_manips", 'toolbar-view-manips.xml',
        _("View (Alternative/Secondary)")),
    ("toolbar1_view_resets", 'toolbar-view-resets.xml',
        _("View (Resetting)")),
]


## Class definitions

class ToolbarManager (object):
    """Manager for toolbars, currently just the main one.

    The main toolbar, /toolbar1, contains a menu button and quick
    access to the painting tools.
    """

    def __init__(self, draw_window):
        super(ToolbarManager, self).__init__()
        self.draw_window = draw_window
        self.app = draw_window.app
        self.toolbar1_ui_loaded = {}  # {name: merge_id, ...}
        self.init_actions()
        ui_dir = os.path.dirname(os.path.abspath(__file__))
        toolbarpath = os.path.join(ui_dir, FRAMEWORK_XML)
        self.app.ui_manager.add_ui_from_file(toolbarpath)
        self.toolbar1 = self.app.ui_manager.get_widget('/toolbar1')
        self.toolbar1.set_style(Gtk.ToolbarStyle.ICONS)
        self.toolbar1.set_icon_size(widgets.get_toolbar_icon_size())
        self.toolbar1.set_border_width(0)
        self.toolbar1.set_show_arrow(True)
        self.toolbar1.connect(
            "popup-context-menu",
            self.on_toolbar1_popup_context_menu
        )
        self.toolbar1_popup = self.app.ui_manager\
            .get_widget('/toolbar1-settings-menu')
        for item in self.toolbar1:
            if isinstance(item, Gtk.SeparatorToolItem):
                item.set_draw(False)
        self.toolbar2 = self.app.ui_manager.get_widget('/toolbar2')
        self.toolbar2.set_style(Gtk.ToolbarStyle.ICONS)
        self.toolbar2.set_icon_size(widgets.get_toolbar_icon_size())
        self.toolbar2.set_border_width(0)
        self.toolbar2.set_show_arrow(False)
        for toolbar in (self.toolbar1, self.toolbar2):
            styles = toolbar.get_style_context()
            styles.add_class(Gtk.STYLE_CLASS_PRIMARY_TOOLBAR)

        # Merge in UI pieces based on the user's saved preferences
        for action in self.settings_actions:
            name = action.get_property("name")
            active = self.app.preferences["ui.toolbar_items"].get(name, False)
            action.set_active(active)
            action.toggled()

    def init_actions(self):
        ag = self.draw_window.action_group
        actions = []

        self.settings_actions = []
        for name, ui_xml, label in MERGEABLE_XML:
            action = Gtk.ToggleAction.new(name, label, None, None)
            action.connect("toggled", self.on_settings_toggle, ui_xml)
            self.settings_actions.append(action)
        actions += self.settings_actions

        for action in actions:
            ag.add_action(action)

    def on_toolbar1_popup_context_menu(self, toolbar, x, y, button):
        menu = self.toolbar1_popup

        def _posfunc(*a):
            return x, y, True
        time = Gtk.get_current_event_time()
        menu.popup(None, None, _posfunc, None, button, time)

    def on_settings_toggle(self, toggleaction, ui_xml_file):
        name = toggleaction.get_property("name")
        merge_id = self.toolbar1_ui_loaded.get(name, None)
        if toggleaction.get_active():
            self.app.preferences["ui.toolbar_items"][name] = True
            if merge_id is not None:
                return
            ui_dir = os.path.dirname(os.path.abspath(__file__))
            ui_xml_path = os.path.join(ui_dir, ui_xml_file)
            merge_id = self.app.ui_manager.add_ui_from_file(ui_xml_path)
            self.toolbar1_ui_loaded[name] = merge_id
        else:
            self.app.preferences["ui.toolbar_items"][name] = False
            if merge_id is None:
                return
            self.app.ui_manager.remove_ui(merge_id)
            self.toolbar1_ui_loaded.pop(name)



# This file is part of MyPaint.
# Copyright (C) 2020 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import division, print_function

import logging

logger = logging.getLogger(__name__)


def validate(input_value, default_value, input_type, predicate, error_message):
    """ Validate user input, returning a default value on failed validations

    :param input_value: The value to validate
    :param default_value: The default value to return if validation fails
    :param input_type: Type or constructor taking the input value
    :param predicate: A unary truth-value function which is run against the
    output of the type / constructor applied to the input value.
    :param error_message: Error message to log on validation failure. The
    message may contain "{value}" to refer to the input value.
    :return: The converted input value if validation is successful, otherwise
    the given default value.
    """
    try:
        value = input_type(input_value)
        assert predicate(value)
        return value
    except Exception:
        logger.exception()
        logger.error(error_message.format(value=input_value))
        return default_value



# This file is part of MyPaint.
# Copyright (C) 2014-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Workspaces with a central canvas, sidebars and saved layouts"""

## Imports

from __future__ import division, print_function
from warnings import warn
import logging

from lib.gibindings import GObject
from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GLib

from gui.toolstack import ToolStack, ToolStackWindow
from gui.windowing import set_initial_window_position
from lib.observable import event
from . import objfactory

logger = logging.getLogger(__name__)


# Class defs


class Workspace (Gtk.VBox, Gtk.Buildable):
    """Widget housing a central canvas flanked by two sidebar toolstacks

    Workspaces also manage zero or more floating ToolStacks, and can set the
    initial size and position of their own toplevel window.

    Instances of tool widget classes can be constructed, then shown and hdden
    by the workspace programmatically using their GType name and an optional
    sequence of construction parameters as a key.  They should support the
    following Python properties:

    * ``tool_widget_icon_name``: the name of the icon to use.
    * ``tool_widget_title``: the title to display in the tooltip, and in
      floating window titles.
    * ``tool_widget_description``: the description string to show in the
      tooltip.

    and the following methods:

    * ``tool_widget_properties()``: show the properties dialog.
    * ``tool_widget_get_icon_pixbuf(size)``: returns a pixbuf icon for a
      particular pixel size. This is used in preference to the icon name.

    Defaults will be used if these properties and methods aren't defined, but
    the defaults are unlikely to be useful.

    The entire layout of a Workspace, including the toplevel window of the
    Workspace itself, can be dumped to and built from structures containing
    only simple Python types.  Widgets can be hidden by the user by clicking on
    a tab group close button within a ToolStack widget, moved around between
    stacks, or snapped out of stacks into new floating windows.

    Workspaces observe their toplevel window, and automatically hide their
    sidebars and floating windows in fullscreen. Auto-hidden elements are
    revealed temporarily when the pointer moves near them. Autohide can be
    toggled off or on, and the setting is retained in the layout definition.

    Workspaces can also manage the visibility of a header and a footer bar
    widget in the same manner: these bars are assumed to be packed above or
    below the Workspace respectively.

    """

    ## Class vars

    #: How near the pointer needs to be to a window edge or a hidden window to
    #: automatically reveal it when autohide is enabled in fullscreen.
    AUTOHIDE_REVEAL_BORDER = 50

    #: Time in milliseconds to wait before revealing UI elements when pointer
    #: is near a window edge.  Does not affect floating hidden windows.
    AUTOHIDE_REVEAL_TIMEOUT = 500

    #: Time in milliseconds to wait before hiding UI elements when autohide is
    #: enabled in fullscreen.
    AUTOHIDE_TIMEOUT = 800

    #: Mask for all buttons.
    #: Used to prevent autohide reveals when a pointer button is pressed
    #: down. Prevents a possible source of distraction when the user is
    #: drawing.
    _ALL_BUTTONS_MASK = (
        Gdk.ModifierType.BUTTON1_MASK | Gdk.ModifierType.BUTTON2_MASK |
        Gdk.ModifierType.BUTTON3_MASK | Gdk.ModifierType.BUTTON4_MASK |
        Gdk.ModifierType.BUTTON5_MASK
    )

    # Edges the pointer can bump: used for autohide reveals

    _EDGE_NONE = 0x00
    _EDGE_LEFT = 0x01
    _EDGE_RIGHT = 0x02
    _EDGE_TOP = 0x04
    _EDGE_BOTTOM = 0x08

    ## GObject integration (type name, properties)

    __gtype_name__ = 'MyPaintWorkspace'

    #: Title suffix property for floating windows.
    floating_window_title_suffix = GObject.Property(
        type=str,
        flags=GObject.ParamFlags.READWRITE,
        nick='Floating window title suffix',
        blurb='The suffix to append to floating windows: typically a '
              'hyphen followed by the application name.',
        default=None
    )

    #: Title separator property for floating windows.
    floating_window_title_separator = GObject.Property(
        type=str,
        flags=GObject.ParamFlags.READWRITE,
        nick='Floating window title separator',
        blurb='String used to separate the names of tools in a '
              'floating window. By default, a comma is used.',
        default=", "
    )

    #: Header bar widget, to be hidden when entering fullscreen mode. This
    #: widget should be packed externally to the workspace, and to its top.
    header_bar = GObject.Property(
        type=Gtk.Widget,
        flags=GObject.ParamFlags.READWRITE,
        nick='Header bar widget',
        blurb="External Menubar/toolbar widget to be hidden when "
              "entering fullscreen mode, and re-shown when leaving "
              "it. The pointer position is also used for reveals and "
              "hides in fullscreen.",
        default=None
    )

    #: Footer bar widget, to be hidden when entering fullscreen mode. This
    #: widget should be packed externally to the workspace, and to its bottom.
    footer_bar = GObject.Property(
        type=Gtk.Widget,
        flags=GObject.ParamFlags.READWRITE,
        nick='Footer bar widget',
        blurb="External footer bar widget to be hidden when entering "
              "fullscreen mode, and re-shown when leaving it. The "
              "pointer position is also used for reveals and hides "
              "in fullscreen.",
        default=None
    )

    def __init__(self):
        """Initializes, with a placeholder canvas widget and no tool widgets"""
        Gtk.VBox.__init__(self)
        # Sidebar stacks
        self._lstack = lstack = ToolStack(self)
        self._rstack = rstack = ToolStack(self)
        lscrolls = Gtk.ScrolledWindow()
        rscrolls = Gtk.ScrolledWindow()
        self._lscrolls = lscrolls
        self._rscrolls = rscrolls
        lscrolls.add(lstack)
        rscrolls.add(rstack)
        for scrolls in [lscrolls, rscrolls]:
            try:
                # Fix 3.16+'s mystery meat: don't use fading scrollbars.
                # Corrects a terrible upstream UX decision which makes
                # the scrollbar cover our widgets.
                scrolls.set_overlay_scrolling(False)
            except AttributeError:
                pass
            scrolls.set_shadow_type(Gtk.ShadowType.IN)
            scrolls.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self._lpaned = lpaned = Gtk.HPaned()
        self._rpaned = rpaned = Gtk.HPaned()
        for stack, paned in [(lstack, lpaned), (rstack, rpaned)]:
            stack.connect("hide", self._sidebar_stack_hide_cb, paned)
            try:
                # Fix 3.16+'s mystery meat: don't hide the divider.
                # Corrects a terrible upstream UX decision which makes
                # things hugely worse for touchscreen and tablet users.
                paned.set_wide_handle(True)
            except AttributeError:
                pass
        # Canvas scrolls. The canvas isn't scrollable yet, but use the same
        # class as the right and left sidebars so that the shadows match
        # in all themes.
        cscrolls = Gtk.ScrolledWindow()
        cscrolls.set_shadow_type(Gtk.ShadowType.IN)
        cscrolls.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
        self._canvas_scrolls = cscrolls
        # Sidebar packing
        lpaned.pack1(lscrolls, resize=False, shrink=False)
        lpaned.pack2(rpaned, resize=True, shrink=False)
        rpaned.pack2(rscrolls, resize=False, shrink=False)
        rpaned.pack1(cscrolls, resize=True, shrink=False)
        self.pack_start(lpaned, True, True, 0)
        # Autohide
        self._autohide_enabled = True
        self._autohide_timeout = None
        self._autoreveal_timeout = []
        # Window tracking
        self._floating = set()
        self._toplevel_pos = dict()
        self._save_toplevel_pos_timeout = None
        self._is_fullscreen = False
        self._is_maximized = False
        self._fs_event_handlers = []
        # Initial layout happens in several phases
        self._initial_layout = None
        self.connect("realize", self._realize_cb)
        self.connect("map", self._map_cb)
        # Tool widget cache and factory
        self._tool_widgets = objfactory.ObjFactory(gtype=Gtk.Widget)
        self._tool_widgets.object_rebadged += self._tool_widget_rebadged

    ## GtkBuildable implementation (pre-realize)

    def do_add_child(self, builder, child, type_):
        """Adds a child as the canvas: gtk_buildable_add_child() impl."""
        self.set_canvas(child)

    ## Setup from layout descriptions (pre-realize)

    def build_from_layout(self, layout):
        """Builds the workspace from a definition dict.

        :param layout: a layout definition
        :type layout: dict

        In order to have any effect, this must be called before the workspace
        widget is realized, but after it has been packed into its toplevel
        window. Keys and values in the dict are as follows:

        * position: an initial window position dict for the toplevel
          window. See `set_initial_window_position()`.
        * left_sidebar, right_sidebar: `ToolStack` definition lists.
          See `Toolstack.build_from_layout()`.
        * floating: a list of floating window definitions. Each element
          is a dict with the following keys:
          - contents: a `ToolStack` definition dict: see above.
          - position: an initial window position dict: see above.
        * autohide: whether autohide is enabled when fullscreening.
        * fullsceen: whether to start in fullscreen mode.
        * maximized: whether to start maximized.

        See also `get_layout()`.

        """
        toplevel_win = self.get_toplevel()
        assert toplevel_win is not None
        assert toplevel_win is not self
        assert not toplevel_win.get_visible()
        # Set initial position and fullscreen state
        toplevel_pos = layout.get("position", None)
        if toplevel_pos:
            set_initial_window_position(toplevel_win, toplevel_pos)
        if layout.get("fullscreen", False):
            toplevel_win.fullscreen()
            GLib.idle_add(lambda *a: toplevel_win.fullscreen())
        elif layout.get("maximized", False):
            toplevel_win.maximize()
            GLib.idle_add(lambda *a: toplevel_win.maximize())
        toplevel_win.connect("window-state-event",
                             self._toplevel_window_state_event_cb)
        self.autohide_enabled = layout.get("autohide", True)
        self._initial_layout = layout

    def get_layout(self):
        """Returns a layout definition dict for the workspace

        This should be called before the toplevel window is fully destroyed,
        or the dicts representing the tool stacks will be empty.

        """
        llayout = self._lstack.get_layout()
        rlayout = self._rstack.get_layout()
        float_layouts = [w.get_layout() for w in self._floating]
        return dict(left_sidebar=llayout, right_sidebar=rlayout,
                    floating=float_layouts, position=self._toplevel_pos,
                    autohide=self._autohide_enabled,
                    fullscreen=self._is_fullscreen,
                    maximized=self._is_maximized)

    ## Initial layout (pre/post-realize)

    def _realize_cb(self, widget):
        """Kick off the deferred layout code when the widget is realized"""

        # Set up monitoring of the toplevel's size changes.
        toplevel = self.get_toplevel()
        toplevel.connect("configure-event", self._toplevel_configure_cb)

        # Do the initial layout
        layout = self._initial_layout
        if layout is None:
            return
        llayout = layout.get("left_sidebar", {})
        if llayout:
            logger.debug("Left sidebar: building from saved layout...")
            num_added_l = self._lstack.build_from_layout(llayout)
            logger.debug("Left sidebar: added %d group(s)", num_added_l)
        rlayout = layout.get("right_sidebar", {})
        if rlayout:
            logger.debug("Right sidebar: building from saved layout...")
            num_added_r = self._rstack.build_from_layout(rlayout)
            logger.debug("Right sidebar: added %d group(s)", num_added_r)
        # Floating windows
        for fi, flayout in enumerate(layout.get("floating", [])):
            logger.debug(
                "Floating window %d: building from saved layout...",
                fi,
            )
            win = ToolStackWindow(self)
            self.floating_window_created(win)
            num_added_f = win.build_from_layout(flayout)
            logger.debug(
                "Floating window %d: added %d group(s)",
                fi,
                num_added_f,
            )
            # The populated ones are only revealed after their
            # floating_window_created have had a chance to run.
            if num_added_f > 0:
                self._floating.add(win)
                GLib.idle_add(win.show_all)
            else:
                logger.warning(
                    "Floating window %d is initially unpopulated. "
                    "Destroying it.",
                    fi,
                )
                GLib.idle_add(win.destroy)

    def _map_cb(self, widget):
        assert self.get_realized()
        logger.debug("Completing layout (mapped)")
        GLib.idle_add(self._complete_initial_layout)

    def _complete_initial_layout(self):
        """Finish initial layout; called after toplevel win is positioned"""
        # Restore saved widths for the sidebar
        layout = self._initial_layout
        if layout is not None:
            left_width = layout.get("left_sidebar", {}).get("w", None)
            if left_width is not None:
                self.set_left_sidebar_width(left_width)
            right_width = layout.get("right_sidebar", {}).get("w", None)
            if right_width is not None:
                self.set_right_sidebar_width(right_width)
        # Sidebar stacks are initially shown, but their contents are not
        # because building from layout was deferred. Hide empties and issue
        # show_all()s.
        if self._lstack.is_empty():
            self._lstack.hide()
        else:
            self._lstack.show_all()
        if self._rstack.is_empty():
            self._rstack.hide()
        else:
            self._rstack.show_all()
        # Toolstacks are responsible for their groups' positions
        for stack in self._get_tool_stacks():
            stack._complete_initial_layout()

    ## Canvas widget

    def set_canvas(self, widget):
        """Canvas widget (setter)"""
        assert self.get_canvas() is None
        widget = self._canvas_scrolls.add(widget)

    def get_canvas(self):
        """Canvas widget (getter)"""
        widget = self._canvas_scrolls.get_child()
        return widget

    def _update_canvas_scrolledwindow(self):
        """Update the canvas ScrolledWindow's border."""
        parent = self._canvas_scrolls
        if not self._is_fullscreen:
            parent.set_shadow_type(Gtk.ShadowType.NONE)
        else:
            parent.set_shadow_type(Gtk.ShadowType.IN)
        # TODO: this should really be done with CSS now.

    ## Tool widgets

    def reveal_tool_widget(self, tool_gtypename, tool_params):
        """Show and present a widget.

        This is a keyboard-friendly alternative to the add/remove toggle
        actions. If the widget is not currently added, it is added.
        Otherwise it is revealed, meaning that its enclosing stack is
        presented to the user and the widget's tab is set as the current
        one.

        """
        if not self.get_tool_widget_added(tool_gtypename, tool_params):
            logger.debug(
                "Reveal: %r %r not in UI, attempting to add it",
                tool_gtypename, tool_params,
            )
            self.add_tool_widget(tool_gtypename, tool_params)
            return
        logger.debug(
            "Reveal: %r %r already in UI, finding and revealing its stack",
            tool_gtypename, tool_params,
        )
        assert self._tool_widgets.cache_has(tool_gtypename, *tool_params)
        widget = self._tool_widgets.get(tool_gtypename, *tool_params)
        assert widget is not None
        stack = self._get_tool_widget_stack(widget)
        assert stack and isinstance(stack, ToolStack)
        stack.reveal_tool_widget(widget)

    def add_tool_widget(self, tool_gtypename, tool_params):
        """Shows a tool widget identified by GType name and construct params

        :param tool_gtypename: GType system name for the new widget's class
        :param tool_params: parameters for the class's Python constructor

        The widget will be created if it doesn't already exist. It will be
        added to the first stack available. Existing floating windows will be
        favoured over the sidebars; if there are no stacks visible, a sidebar
        will be made visible to receive the new widget.

        """
        logger.debug("Adding %r %r", tool_gtypename, tool_params)
        # Attempt to get the widget, potentially creating it here
        try:
            widget = self._tool_widgets.get(tool_gtypename, *tool_params)
        except objfactory.ConstructError as ex:
            logger.error("add_tool_widget: %s", ex.message)
            return
        # Inject it into a suitable ToolStack
        stack = None
        if widget.get_parent() is not None:
            logger.debug("Existing %r is already added", widget)
            stack = self._get_tool_widget_stack(widget)
        else:
            logger.debug("Showing %r, which is currently hidden", widget)
            maxpages = 1
            added = False
            stack = None
            while maxpages < 100 and not added:
                for stack in self._get_tool_stacks():
                    if stack.add_tool_widget(widget, maxnotebooks=3,
                                             maxpages=maxpages):
                        added = True
                        break
                maxpages += 1
            if not added:
                logger.error("Can't find space for %r in any stack", widget)
                return
        # Reveal the widget's ToolStack
        assert stack and isinstance(stack, ToolStack)
        stack.reveal_tool_widget(widget)

    def remove_tool_widget(self, tool_gtypename, tool_params):
        """Removes a tool widget by typename+params

        This hides the widget and orphans it from the widget hierarchy, but a
        reference to it is kept in the Workspace's internal cache. Further
        calls to add_tool_widget() will use the cached object.

        :param tool_gtypename: GType system name for the widget's class
        :param tool_params: construct params further identifying the widget
        :returns: whether the widget was found and hidden
        :rtype: bool

        """
        # First, does it even exist?
        if not self._tool_widgets.cache_has(tool_gtypename, *tool_params):
            return False
        # Can't remove anything that's already removed
        widget = self._tool_widgets.get(tool_gtypename, *tool_params)
        if widget.get_parent() is None:
            return False
        # The widget should exist in a known stack; find and remove
        for stack in self._get_tool_stacks():
            if stack.remove_tool_widget(widget):
                return True
        # Should never happen...
        warn("Asked to remove a visible widget, but it wasn't in any stack",
             RuntimeWarning)
        return False

    def get_tool_widget_added(self, gtype_name, params):
        """Returns whether a tool widget is currently in the widget tree"""
        # Nonexistent objects are not parented or showing
        if not self._tool_widgets.cache_has(gtype_name, *params):
            return False
        # Otherwise, just test whether it's in a widget tree
        widget = self._tool_widgets.get(gtype_name, *params)
        return widget.get_parent() is not None

    def _get_tool_widget_stack(self, widget):
        """Gets the ToolStack a widget is currently parented underneath."""
        stack = None
        if widget.get_parent() is not None:
            stack = widget.get_parent()
            while stack and not isinstance(stack, ToolStack):
                stack = stack.get_parent()
        return stack

    def update_tool_widget_params(self, tool_gtypename,
                                  old_params, new_params):
        """Update the construction params of a tool widget

        :param tool_gtypename: GType system name for the widget's class
        :param old_params: old parameters for the class's Python constructor
        :param new_params: new parameters for the class's Python constructor

        If an object has changed so that its construction parameters must be
        update, this method should be called to keep track of its identity
        within the workspace.  This method will not show or hide it: its
        current state remains the same.

        See also `update_tool_widget_ui()`.

        """
        # If it doesn't exist yet, updating what is effectively
        # a cache key used for accessing it makes no sense.
        if not self._tool_widgets.cache_has(tool_gtypename, *old_params):
            return
        # Update the params of an existing object.
        widget = self._tool_widgets.get(tool_gtypename, *old_params)
        if old_params == new_params:
            logger.devug("No construct params update needed for %r", widget)
        else:
            logger.debug("Updating construct params for %r", widget)
            self._tool_widgets.rebadge(widget, new_params)

    def update_tool_widget_ui(self, gtype_name, params):
        """Updates tooltips and tab labels for a specified tool widget

        Calling this method causes the workspace to re-read the tool widget's
        tab label, window title, and tooltip properties, and update the
        display.  Use it when things like the icon pixbuf have changed.  It's
        not necessary to call this after `update_tool_widget_params()` has been
        used: that's handled with an event.

        """
        if not self.get_tool_widget_added(gtype_name, params):
            return
        widget = self._tool_widgets.get(gtype_name, *params)
        logger.debug("Updating workspace UI widgets for %r", widget)
        self._update_tool_widget_ui(widget)

    ## Tool widget events

    @event
    def tool_widget_added(self, widget):
        """Event: tool widget added"""

    @event
    def tool_widget_removed(self, widget):
        """Event: tool widget removed either by the user or programmatically"""

    @event
    def floating_window_created(self, toplevel):
        """Event: a floating window was created to house a toolstack."""

    @event
    def floating_window_destroyed(self, toplevel):
        """Event: a floating window was just `destroy()`ed."""

    ## Sidebar toolstack width

    def set_right_sidebar_width(self, width):
        """Sets the width of the right sidebar toolstack
        """
        if self._rstack.is_empty():
            return
        width = max(width, 100)
        handle_size = GObject.Value()
        handle_size.init(int)
        self._rpaned.style_get_property("handle-size", handle_size)
        position = self._rpaned.get_allocated_width()
        position -= width
        position -= handle_size.get_int()
        self._rpaned.set_position(position)

    def set_left_sidebar_width(self, width):
        """Sets the width of the left sidebar toolstack
        """
        if self._lstack.is_empty():
            return
        width = max(width, 100)
        self._lpaned.set_position(width)

    ## Position saving (toplevel window)

    def _toplevel_configure_cb(self, toplevel, event):
        """Record the toplevel window's position ("configure-event" callback)
        """
        # Avoid saving fullscreen positions. The timeout is a bit of hack, but
        # it's necessary because the state change event and the configure event
        # when fullscreening don't have a sensible order.
        w, h = event.width, event.height
        srcid = self._save_toplevel_pos_timeout
        if srcid:
            GLib.source_remove(srcid)
        srcid = GLib.timeout_add(
            250,
            self._save_toplevel_pos_timeout_cb,
            w, h,
        )
        self._save_toplevel_pos_timeout = srcid

    def _save_toplevel_pos_timeout_cb(self, w, h):
        """Toplevel window position recording (post-"configure-event" oneshot)

        Saves the (x,y) from the frame and the (w,h) from the configure event:
        the combination can be used as an initial size and position next time.

        """
        self._save_toplevel_pos_timeout = None
        if self._is_fullscreen or self._is_maximized:
            return False
        toplevel = self.get_toplevel()
        gdk_win = toplevel.get_window()
        extents = gdk_win.get_frame_extents()
        x = max(0, extents.x)
        y = max(0, extents.y)
        pos = dict(x=x, y=y, w=w, h=h)
        self._toplevel_pos = pos
        return False

    ## Toolstack order for searching, tool insertion etc.

    def _get_tool_stacks(self):
        """Yields all known ToolStacks, in floating-first order.
        """
        for win in self._floating:
            yield win.stack
        yield self._rstack
        yield self._lstack

    ## Tool widget tab dragging (event callbacks)

    def _tool_tab_drag_begin_cb(self):
        """Shows all possible drag targets at the start of a tool tab drag

        Ensures that all the known toolstacks are visible to receive the drag,
        even those which are empty or hidden due to fullscreening. Called by
        stack notebooks in this workspace when tab drags start.

        """
        # First cancel any pending hides.
        if self._is_fullscreen:
            self._cancel_autohide_timeout()
        # Ensure the left and right stacks are visible at the beginning of
        # a tab drag even if empty so that the user can drop a tab there.
        for stack in (self._lstack, self._rstack):
            scrolls = stack.get_parent().get_parent()
            empty = stack.is_empty()
            visible = stack.get_visible() and scrolls.get_visible()
            if (empty and not visible) or (not empty and not visible):
                scrolls = stack.get_parent().get_parent()
                scrolls.show_all()

    def _tool_tab_drag_end_cb(self):
        """Hides empty toolstacks at the end of a tool tab drag

        Called by stack notebooks in this workspace when tab drags finish.

        """
        for stack in (self._lstack, self._rstack):
            scrolls = stack.get_parent().get_parent()
            empty = stack.is_empty()
            visible = stack.get_visible() and scrolls.get_visible()
            if empty and visible:
                stack.hide()

    def _sidebar_stack_hide_cb(self, stack, paned):
        """Resets sidebar sizes when they're emptied (sidebar "hide" callback)

        If the hide is due to the sidebar stack having been emptied out,
        resetting the size means that it'll show at its placeholder's size when
        it's next shown by the drag-start handler: this makes a narrower but
        less intrusive target for the drag.

        """
        if stack.is_empty():
            paned.set_position(-1)
        # Hide the parent GtkScrolledWindow too, allowing the paned to not
        # show the sidebar pane at all - GtkScrolledWindows have an allocation
        # even if they're empty, assuming they have scrollbars.
        scrolls = stack.get_parent().get_parent()
        scrolls.hide()

    ## Fullscreen (event callbacks)

    def _toplevel_window_state_event_cb(self, toplevel, event):
        """Handle transitions between fullscreen and windowed."""
        if event.changed_mask & Gdk.WindowState.FULLSCREEN:
            fullscreen = event.new_window_state & Gdk.WindowState.FULLSCREEN
            if fullscreen:
                if self.autohide_enabled:
                    self._connect_autohide_events()
                    self._start_autohide_timeout()
                # Showing the floating windows makes an initial fullscreen
                # look a little nicer.
                for floating in self._floating:
                    floating.show_all()
            else:
                self._disconnect_autohide_events()
                self._show_autohide_widgets()
            self._is_fullscreen = bool(fullscreen)
            self._update_canvas_scrolledwindow()
        if event.changed_mask & Gdk.WindowState.MAXIMIZED:
            maximized = event.new_window_state & Gdk.WindowState.MAXIMIZED
            self._is_maximized = bool(maximized)

    ## Autohide flag

    def get_autohide_enabled(self):
        """Auto-hide is enabled in fullscreen (getter)"""
        return self._autohide_enabled

    def set_autohide_enabled(self, autohide_enabled):
        """Auto-hide is enabled in fullscreen (setter)"""
        if self._is_fullscreen:
            if autohide_enabled:
                self._connect_autohide_events()
                self._hide_autohide_widgets()
            else:
                self._disconnect_autohide_events()
                self._show_autohide_widgets()
        self._autohide_enabled = bool(autohide_enabled)

    autohide_enabled = property(get_autohide_enabled, set_autohide_enabled)

    def _hide_autohide_widgets(self):
        """Hides all auto-hiding widgets immediately"""
        if not self._is_fullscreen:
            return
        self._cancel_autohide_timeout()
        display = self.get_display()
        if display.pointer_is_grabbed():
            logger.warning("Pointer grabbed: not auto-hiding")
            return
        ah_widgets = self._get_autohide_widgets()
        logger.debug("Hiding %d autohide widget(s)", len(ah_widgets))
        for widget in ah_widgets:
            if widget.get_visible():
                widget.hide()

    def _show_autohide_widgets(self):
        """Shows all auto-hiding widgets immediately"""
        self._cancel_autohide_timeout()
        ah_widgets = self._get_autohide_widgets()
        logger.debug("Hiding %d autohide widget(s)", len(ah_widgets))
        for widget in ah_widgets:
            widget.show_all()

    def _get_autohide_widgets(self):
        """List of autohide widgets

        Returns a list of the widgets which should be revealed by edge bumping,
        rollovers, or hidden by the timeout.

        """
        widgets = []
        for stack in [self._lstack, self._rstack]:
            if not stack.is_empty():
                scrolls = stack.get_parent().get_parent()
                widgets.append(scrolls)
        widgets.extend(list(self._floating))
        for bar in [self.header_bar, self.footer_bar]:
            if bar:
                widgets.append(bar)
        return widgets

    ## Autohide mode: auto-hide timer

    def _start_autohide_timeout(self):
        """Start a timer to hide the UI after a brief period of inactivity"""
        if not self._autohide_timeout:
            logger.debug("Starting autohide timeout (%d milliseconds)",
                         self.AUTOHIDE_TIMEOUT)
        else:
            self._cancel_autohide_timeout()
        srcid = GLib.timeout_add(
            self.AUTOHIDE_TIMEOUT,
            self._autohide_timeout_cb,
        )
        self._autohide_timeout = srcid

    def _cancel_autohide_timeout(self):
        """Cancels any pending auto-hide"""
        if not self._autohide_timeout:
            return
        GLib.source_remove(self._autohide_timeout)
        self._autohide_timeout = None

    def _autohide_timeout_cb(self):
        """Hide auto-hide widgets when the auto-hide timer finishes"""
        self._hide_autohide_widgets()
        return False

    # Autohide mode: auto-reveal timer

    def _start_autoreveal_timeout(self, widget):
        """Start a timer to reveal the widget after a brief period
        of edge contact
        """
        if not self._autoreveal_timeout:
            logger.debug("Starting autoreveal timeout (%d milliseconds)",
                         self.AUTOHIDE_REVEAL_TIMEOUT)
        else:
            self._cancel_autoreveal_timeout()
        srcid = GLib.timeout_add(
            self.AUTOHIDE_REVEAL_TIMEOUT,
            self._autoreveal_timeout_cb,
            widget,
        )
        self._autoreveal_timeout.append(srcid)

    def _cancel_autoreveal_timeout(self):
        """Cancels any pending auto-reveal"""
        if not self._autoreveal_timeout:
            return
        for timer in self._autoreveal_timeout:
            GLib.source_remove(timer)
        self._autoreveal_timeout = []

    def _autoreveal_timeout_cb(self, widget):
        """Show widgets when the auto-reveal timer finishes"""
        widget.show_all()
        self._autoreveal_timeout = []
        return False

    ## Autohide mode: event handling on the canvas widget

    def _connect_autohide_events(self):
        """Start listening for autohide events"""
        if self._fs_event_handlers:
            return
        evwidget = self.get_canvas()
        if not evwidget:
            return
        mask = (
            Gdk.EventMask.POINTER_MOTION_HINT_MASK |
            Gdk.EventMask.POINTER_MOTION_MASK |
            Gdk.EventMask.LEAVE_NOTIFY_MASK |
            Gdk.EventMask.ENTER_NOTIFY_MASK
        )
        evwidget.add_events(mask)
        handlers = [("motion-notify-event", self._fs_motion_cb),
                    ("leave-notify-event", self._fs_leave_cb),
                    ("enter-notify-event", self._fs_enter_cb)]
        for event_name, handler_callback in handlers:
            handler_id = evwidget.connect(event_name, handler_callback)
            self._fs_event_handlers.append((evwidget, handler_id))

    def _disconnect_autohide_events(self):
        """Stop listening for autohide events"""
        for evwidget, handler_id in self._fs_event_handlers:
            evwidget.disconnect(handler_id)
        self._fs_event_handlers = []

    def _fs_leave_cb(self, widget, event):
        """Handles leaving the canvas in fullscreen"""
        assert self._is_fullscreen
        # if event.state & self._ALL_BUTTONS_MASK:
        #    # "Starting painting", except not quite.
        #    # Can't use this: there's no way of distinguishing it from
        #    # resizing a floating window. Hiding the window being resized
        #    # breaks window management badly! (Xfce 4.10)
        #    self._cancel_autohide_timeout()
        #    self._hide_autohide_widgets()
        if event.mode == Gdk.CrossingMode.UNGRAB:
            # Finished painting. To appear more consistent with a mouse,
            # restart the hide timer now rather than waiting for a motion
            # event.
            self._start_autohide_timeout()
        elif event.mode == Gdk.CrossingMode.NORMAL:
            # User may be using a sidebar. Leave it open.
            self._cancel_autohide_timeout()
        return False

    def _fs_enter_cb(self, widget, event):
        """Handles entering the canvas in fullscreen"""
        assert self._is_fullscreen
        # If we're safely in the middle, the autohide timer can begin now.
        if not self._get_bumped_edges(widget, event):
            self._start_autohide_timeout()
        return False

    def _fs_motion_cb(self, widget, event):
        """Handles edge bumping and other rollovers in fullscreen mode"""
        assert self._is_fullscreen
        # Firstly, if the user appears to be drawing, be as stable as we can.
        if event.state & self._ALL_BUTTONS_MASK:
            self._cancel_autohide_timeout()
            self._cancel_autoreveal_timeout()
            return False
        # Floating window rollovers
        show_floating = False
        for win in self._floating:
            if win.get_visible():
                continue
            x, y = event.x_root, event.y_root
            b = self.AUTOHIDE_REVEAL_BORDER
            if win.contains_point(x, y, b=b):
                show_floating = True
        if show_floating:
            for win in self._floating:
                win.show_all()
            self._cancel_autohide_timeout()
            return False
        # Edge bumping
        # Bump the mouse into the edge of the screen to get back the stuff
        # that was hidden there, similar to media players etc.
        edges = self._get_bumped_edges(widget, event)
        if not edges:
            self._start_autohide_timeout()
            self._cancel_autoreveal_timeout()
            return False
        if edges & self._EDGE_TOP and self.header_bar:
            self._start_autoreveal_timeout(self.header_bar)
        if edges & self._EDGE_BOTTOM and self.footer_bar:
            self._start_autoreveal_timeout(self.footer_bar)
        if edges & self._EDGE_LEFT and not self._lstack.is_empty():
            self._start_autoreveal_timeout(self._lscrolls)
        if edges & self._EDGE_RIGHT and not self._rstack.is_empty():
            self._start_autoreveal_timeout(self._rscrolls)

    @classmethod
    def _get_bumped_edges(cls, widget, event):
        # Returns a bitmask of the edges bumped by the pointer.
        alloc = widget.get_allocation()
        w, h = alloc.width, alloc.height
        x, y = event.x, event.y
        b = cls.AUTOHIDE_REVEAL_BORDER
        if not (x < b or x > w-b or y < b or y > h-b):
            return cls._EDGE_NONE
        edges = cls._EDGE_NONE
        if y < b or (y < 5*b and (x < b or x > w-b)):
            edges |= cls._EDGE_TOP
        if y > h - b:
            edges |= cls._EDGE_BOTTOM
        if x < b:
            edges |= cls._EDGE_LEFT
        if x > w - b:
            edges |= cls._EDGE_RIGHT
        return edges

    ## Tool widget tab & title updates

    def _tool_widget_rebadged(self, factory, product, old_params, new_params):
        """Internal: update UI elements when the ID of a tool widget changes

        For parameterized ones like the brush group tool widget, the tooltip
        and titlebar are dependent on the identity strings and must be update
        when the tab is renamed.

        """
        self._update_tool_widget_ui(product)

    def _update_tool_widget_ui(self, widget):
        """Internal: update UI elements for a known descendent tool widget"""
        page = widget.get_parent()
        notebook = page.get_parent()
        notebook.update_tool_widget_ui(widget)


## Module testing (interactive, but fairly minimal)

def _test():
    logging.basicConfig(level=logging.DEBUG)
    import os
    import sys

    class _TestLabel (Gtk.Label):
        __gtype_name__ = 'TestLabel'
        tool_widget_icon_name = 'gtk-ok'
        tool_widget_description = "Just a test widget"

        def __init__(self, text):
            Gtk.Label.__init__(self, text)
            self.set_size_request(200, 150)

    class _TestSpinner (Gtk.Spinner):
        __gtype_name__ = "TestSpinner"
        tool_widget_icon_name = 'gtk-cancel'
        tool_widget_description = "Spinner test"

        def __init__(self):
            Gtk.Spinner.__init__(self)
            self.set_size_request(150, 150)
            self.set_property("active", True)

    def _tool_shown_cb(*a):
        logger.debug("TOOL-SHOWN %r", a)

    def _tool_hidden_cb(*a):
        logger.debug("TOOL-HIDDEN %r", a)

    def _floating_window_created(*a):
        logger.debug("FLOATING-WINDOW-CREATED %r", a)
    workspace = Workspace()
    workspace.floating_window_title_suffix = u" - Test"
    button = Gtk.Button(label="Click to close this demo")
    frame = Gtk.Frame()
    frame.add(button)
    frame.set_shadow_type(Gtk.ShadowType.IN)
    workspace.set_canvas(frame)
    window = Gtk.Window()
    window.add(workspace)
    window.set_title(os.path.basename(sys.argv[0]))
    workspace.set_size_request(600, 400)
    workspace.tool_widget_added += _tool_shown_cb
    workspace.tool_widget_removed += _tool_hidden_cb
    workspace.floating_window_created += _floating_window_created
    workspace.build_from_layout({
        'position': {'x': 100, 'y': 75, 'h': -100, 'w': -100},
        'floating': [{
            'position': {'y': -100, 'h': 189, 'w': 152, 'x': -200},
            'contents': {
                'groups': [{
                    'tools': [('TestLabel', "1"), ('TestLabel', "2")],
                }],
            }}],
        'right_sidebar': {
            'w': 400,
            'groups': [{
                'tools': [('TestSpinner',), ("TestLabel", "3")],
            }],
        },
        'left_sidebar': {
            'w': 250,
            'groups': [{
                'tools': [('TestLabel', "4"), ('TestLabel', "5")],
            }],
        },
        'maximized': False,
        'fullscreen': True,
    })
    window.show_all()

    def _quit_cb(*a):
        logger.info("Demo quit, workspace dump follows")
        print(workspace.get_layout())
        Gtk.main_quit()
    window.connect("destroy", _quit_cb)
    button.connect("clicked", _quit_cb)
    Gtk.main()


if __name__ == '__main__':
    _test()



# -*- coding: utf-8 -*-
# This file is part of MyPaint.
# Copyright (C) 2007-2014 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2015 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import division, print_function

import os
import sys
import time
import tempfile
import subprocess
import shutil
import logging

from lib.gibindings import GLib
from lib.gibindings import Gtk

import lib.fileutils


logger = logging.getLogger(__name__)


# Copied from Python 3.5's distutils/spawn.py

def find_executable(executable, path=None):
    """Tries to find 'executable' in the directories listed in 'path'.

    A string listing directories separated by 'os.pathsep'; defaults to
    os.environ['PATH'].  Returns the complete filename or None if not found.
    """
    if path is None:
        path = os.environ['PATH']

    paths = path.split(os.pathsep)
    base, ext = os.path.splitext(executable)

    if (sys.platform == 'win32') and (ext != '.exe'):
        executable = executable + '.exe'

    if not os.path.isfile(executable):
        for p in paths:
            f = os.path.join(p, executable)
            if os.path.isfile(f):
                # the file exists, we have a shot at spawn working
                return f
        return None
    else:
        return executable


class Profiler (object):
    """Handles profiling state for the main app.

    The profiler's output is written to a tempdir, which is shown to the
    user in their file browser once processing is complete. Ideally
    you'll have gprof2dot.py and graphviz installed, so you can compare
    pretty PNG files directly. However even in the absence of these
    tools, you can copy the output .pstats files elsewhere and run
    gprof2dot.py and dot on them manually.

    The tempdir is deleted when the application exits normally.

    """

    GPROF2DOT = ["gprof2dot.py", "-f", "pstats"]
    DOT2PNG = ["dot", "-Tpng"]

    if find_executable("gprof2dot"):
        GPROF2DOT = ["gprof2dot", "-f", "pstats"]

    def __init__(self):
        super(Profiler, self).__init__()
        self.profiler_active = False
        self.profile_num = 0
        self.__temp_dir = None

    def toggle_profiling(self):
        """Starts profiling if not running, or stops it & shows results."""
        if self.profiler_active:
            self.profiler_active = False
        else:
            GLib.idle_add(self._do_profiling)

    @property
    def _tempdir(self):
        td = self.__temp_dir
        if not td:
            td = tempfile.mkdtemp(prefix="mypaint-profile-")
            self.__temp_dir = td
        return td

    def _do_profiling(self):
        """Runs the GTK main loop in the cProfile profiler till stopped."""
        self.profile_num += 1
        basename = "{isotime}-{n}".format(
            isotime = time.strftime("%Y%m%d-%H%M%S"),
            n = self.profile_num,
        )

        import cProfile
        profile = cProfile.Profile()

        self.profiler_active = True
        logger.info('--- GUI Profiling starts ---')
        while self.profiler_active:
            profile.runcall(Gtk.main_iteration_do, False)
            if not Gtk.events_pending():
                time.sleep(0.050)
                # ugly trick to remove "user does nothing" from profile
        logger.info('--- GUI Profiling ends ---')

        pstats_filepath = os.path.join(self._tempdir, basename + ".pstats")
        if os.path.exists(pstats_filepath):
            os.unlink(pstats_filepath)
        profile.dump_stats(pstats_filepath)
        logger.debug('profile written to %r', pstats_filepath)

        try:
            dot_filepath = os.path.join(self._tempdir, basename + ".dot")
            if os.path.exists(dot_filepath):
                os.unlink(dot_filepath)
            cmd = list(self.GPROF2DOT) + ["-o", dot_filepath, pstats_filepath]
            logger.debug("Running %r...", cmd)
            subprocess.check_call(cmd)

            png_filepath = os.path.join(self._tempdir, basename + ".png")
            if os.path.exists(png_filepath):
                os.unlink(png_filepath)
            cmd = list(self.DOT2PNG) + ["-o", png_filepath, dot_filepath]
            logger.debug("Running %r...", cmd)
            subprocess.check_call(cmd)
        except Exception:
            logger.exception(
                "Profiling output post-processing failed."
            )
            logger.info(
                "This is normal if %r and/or graphviz's %r "
                "are not both in your PATH and executable.",
                self.GPROF2DOT[0],
                self.DOT2PNG[0],
            )
        else:
            logger.info("Profiling post-processing succeeded.")
        logger.info(
            "Opening output folder %r with the default directory viewer...",
            self._tempdir,
        )
        lib.fileutils.startfile(self._tempdir)

    def __del__(self):
        self.cleanup()

    def cleanup(self):
        """Cleans up the tempdir associated with the profile object."""
        if self.__temp_dir is None:
            return
        try:
            if os.path.isdir(self.__temp_dir):
                logger.info("Cleaning up %r...", self.__temp_dir)
                shutil.rmtree(self.__temp_dir, ignore_errors=True)
        except Exception:
            logger.exception("Cleanup of %r failed", self.__temp_dir)
        else:
            self.__temp_dir = None



# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2015-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Pythonic wrappers for some standard GLib routines.

MyPaint is strict about using Unicode internally for everything, but
when GLib returns a filename, all hell can break loose (the data isn't
unicode, and may not even be UTF-8). This module works around that.

"""


## Imports

from __future__ import division, print_function
import logging
import sys

from lib.pycompat import PY3
from lib.pycompat import unicode

from lib.gibindings import GLib

logger = logging.getLogger(__name__)


## File path getter functions

def filename_to_unicode(opsysstring):
    """Converts a str representing a filename from GLib to unicode.

    :param bytes opsysstring: a string in the (GLib) encoding for filenames
    :returns: the converted filename
    :rtype: unicode|str

    >>> filename_to_unicode(b'/ascii/only/path') == u'/ascii/only/path'
    True
    >>> filename_to_unicode(None) is None
    True

    This is just a more Pythonic wrapper around g_filename_to_utf8() for
    now, which works around a ton of weird bugs and corner cases with
    the typelib annotations for it. It is intended for cleaning up the
    output of certain GLib functions.

    Currently, if you're using Python 3 and the input is already unicode
    then this function assumes that GLib+GI have already done the work,
    and that the unicode string was correct. You get the same string
    back.

    For Python 2, this accepts only "bytes" string input. If we find a
    corner case where GLib functions return degenerate unicode, we can
    adapt it for that case (those funcs need their own wrappers though).

    """
    if opsysstring is None:
        return None

    # Let's assume that if the string is already unicode under Python 3,
    # then it's already correct.
    if PY3 and isinstance(opsysstring, unicode):
        return opsysstring

    # On Windows, they're always UTF-8 regardless.
    # That's what the docs say.
    if sys.platform == "win32":
        return opsysstring.decode("utf-8")

    # Other systems are dependent in opaque ways on the environment.
    if not isinstance(opsysstring, bytes):
        raise TypeError("Argument must be bytes")
    opsysstring_degenerate_unicode = opsysstring.decode("latin_1")

    # This function's annotation seems to vary quite a bit.
    # See https://github.com/mypaint/mypaint/issues/634
    ustring = None

    # The sensible, modern case! Byte strings in, unicode strings
    # out hopefully, and the C func's arguments are correctly
    # [out]-annotated. It works like this as of...
    #
    # - Python 2.7.14 OR Python 3.6.4
    # - gobject-introspection 1.54.1
    # - glib2 2.54.3
    # - Debian buster/sid amd64 OR MSYS2 MINGW64 on Windows 7 64-bit.

    if ustring is None:
        for s in [opsysstring, opsysstring_degenerate_unicode]:
            try:
                ustring, _bytes_read, _bytes_written \
                    = GLib.filename_to_utf8(s, -1)
                break
            except TypeError:
                pass

    # Try an older, bad typelib's form.
    # This is the case for Ubuntu 14.04 LTS "trusty" (which is ancient,
    # but that's what our current Travis CI solution uses). For the
    # record, this weirdness is applicable to the following combination:
    #
    # - Python 2.7.6 OR Python 3.4.3
    # - gobject-introspection 1.40.0
    # - glib2 2.40.2
    # - Ubuntu 14.04.5 LTS amd64.
    #
    # Of note: the Py3 wrappings are weird in Trusty. Other GLib funcs
    # return bytes, but GLib.filename_to_utf8() expects those degenerate
    # unicode strings. Byte strings will not do. Unusual tastes.

    if ustring is None:
        for s in [opsysstring, opsysstring_degenerate_unicode]:
            try:
                ustring = GLib.filename_to_utf8(s, -1, 0, 0)
                break
            except TypeError:
                pass

    # Congratulations! You found a new bug.

    if ustring is None:
        raise UnicodeDecodeError(
            "New or unknown bugs in g_filename_to_utf8()'s typelib. "
            "Failed to convert %r. Please tell the developers about this."
            % (opsysstring,)
        )

    # Python2's wrappers tended to do this.
    # I suspect it's reasonable to convert for all, now that we're
    # reasonably sure that the data would be utf-8.

    if isinstance(ustring, bytes):
        ustring = ustring.decode("utf-8")

    return ustring


def get_user_config_dir():
    """Like g_get_user_config_dir(), but always unicode"""
    d_fs = GLib.get_user_config_dir()
    return filename_to_unicode(d_fs)


def get_user_data_dir():
    """Like g_get_user_data_dir(), but always unicode"""
    d_fs = GLib.get_user_data_dir()
    return filename_to_unicode(d_fs)


def get_user_cache_dir():
    """Like g_get_user_cache_dir(), but always unicode"""
    d_fs = GLib.get_user_cache_dir()
    return filename_to_unicode(d_fs)


def get_user_special_dir(d_id):
    """Like g_get_user_special_dir(), but always unicode"""
    d_fs = GLib.get_user_special_dir(d_id)
    return filename_to_unicode(d_fs)


## First-import cache forcing

def init_user_dir_caches():
    """Caches the GLib user directories

    >>> init_user_dir_caches()

    The first time this module is imported is from a particular point in
    the launch script, after all the i18n setup is done and before
    lib.mypaintlib is imported. If they're not cached up-front in this
    manner, get_user_config_dir() & friends may return literal "?"s in
    place of non-ASCII characters (Windows systems with non-ASCII user
    profile dirs are known to trigger this).

    The debugging prints may be useful too.

    """
    logger.debug("Init g_get_user_config_dir(): %r", get_user_config_dir())
    logger.debug("Init g_get_user_data_dir(): %r", get_user_data_dir())
    logger.debug("Init g_get_user_cache_dir(): %r", get_user_cache_dir())
    # It doesn't matter if some of these are None
    for i in range(GLib.UserDirectory.N_DIRECTORIES):
        k = GLib.UserDirectory(i)
        logger.debug(
            "Init g_get_user_special_dir(%s): %r",
            k.value_name,
            get_user_special_dir(k),
        )


## Filename <-> URI conversion


def filename_to_uri(abspath, hostname=None):
    """More Pythonic & stable g_filename_to_uri(), with OS workarounds.

    >>> import os.path
    >>> relpath = os.path.join(u'tmp', u'smile (\u263a).ora')
    >>> abspath = os.path.abspath(relpath)
    >>> uri = filename_to_uri(abspath)
    >>> isinstance(uri, str)
    True
    >>> uri.endswith('/tmp/smile%20(%E2%98%BA).ora')
    True
    >>> uri.startswith('file:///')
    True

    """
    if hostname:
        raise ValueError("Only NULL hostnames are supported")
    # GLib.filename_to_uri is *present* on Windows, for both i686 and
    # x64_64 (MSYS2 builds), however attempting to *call* it on the
    # 64-bit build results in
    # "Error: g-invoke-error-quark: Could not locate g_filename_to_uri"
    # as reported in https://github.com/mypaint/mypaint/issues/374
    # Use the _utf8 variant instead on platforms where it exists.
    try:
        g_filename_to_uri = GLib.filename_to_uri_utf8
    except AttributeError:
        g_filename_to_uri = GLib.filename_to_uri
    hostname = ""
    return g_filename_to_uri(abspath, hostname)


def filename_from_uri(uri):
    """More Pythonic & stable g_filename_from_uri(), with OS workarounds.

    >>> import os.path
    >>> relpath = os.path.join(u'tmp', u'smile (\u263a).ora')
    >>> abspath1 = os.path.abspath(relpath)
    >>> uri = filename_to_uri(abspath1)
    >>> abspath2, hostname = filename_from_uri(uri)
    >>> if PY3:
    ...     unicode = str
    >>> isinstance(abspath2, unicode)
    True
    >>> abspath2.replace('\\\\', "/") == abspath1.replace('\\\\', "/")
    True

    """
    # First find the right g_filename_from_uri.
    # See the note above.
    try:
        g_filename_from_uri = GLib.filename_from_uri_utf8
    except AttributeError:
        g_filename_from_uri = GLib.filename_from_uri
    # But oh joy. We have to support more than one typelib.
    try:
        # Newer GLib typelibs on Linux mark it as a return.
        abspath, hostname = g_filename_from_uri(uri)
    except TypeError:
        # Older GLib typelibs,
        # including the one shipped with Ubuntu Server 12.04 (Travis!)
        # And that windows _utf8 mess still uses it too.
        abspath = g_filename_from_uri(uri, "")
        hostname = None
    assert (not hostname), ("Only URIs without hostnames are supported.")
    return (filename_to_unicode(abspath), None)


## Module testing

def _test():
    import doctest
    doctest.testmod()


if __name__ == '__main__':
    _test()



# This file is part of MyPaint.
# Copyright (C) 2013 by Andrew Chadwick <a.t.chadwick@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Factory for creating custom toolbar and manu items via UIManager"""

from __future__ import division, print_function

from warnings import warn

import gi
from lib.gibindings import Gtk
from lib.gibindings import GObject


class FactoryAction (Gtk.Action):
    """Generic factory action for UI components.

    Define an instance of the factory once for each family of custom proxy
    classes (GtkToolItem, GtkMenuItem etc.) that you need.  Each instance must
    be named after the custom proxy classes the factory is to produce, so the
    definition is best done in a central snippet of GtkBuilder XML.

    For example, if the factory is named ``FooBar``, then its proxy ToolItems
    are expected to have ``__gtype_name__``s of``"MyPaintFooBarToolItem``.

    Creation of proxies can then be done cleanly in the GtkUIManager XML
    definitions by referring to the factory's name as many times as needed.

    """

    #: Predictable name for GtkBuilder XML.
    __gtype_name__ = "MyPaintFactoryAction"

    #: The pattern to use when instantiating a tool item
    TOOL_ITEM_NAME_PATTERN = "MyPaint%sToolItem"

    #: The pattern to use when instantiating a menu item
    MENU_ITEM_NAME_PATTERN = "MyPaint%sMenuItem"

    def __init__(self, *a):
        # GtkAction's own constructor requires params which are all set up by
        # Builder. It warns noisily, so bypass it and invoke its parent
        # class's.
        super(Gtk.Action, self).__init__()

    def do_create_tool_item(self):
        """Returns a new ToolItem

        Invoked by UIManager when it needs a GtkToolItem proxy for a toolbar.

        This method instantiates and returns a new widget from a class named
        after the factory action's own name.  Class lookup is done via GObject:
        see `TOOL_ITEM_NAME_PATTERN` for the ``__gtype_name__`` this method
        will expect.

        """
        gtype_name = self.TOOL_ITEM_NAME_PATTERN % (self.get_name(),)
        tool_item = self._construct(gtype_name)
        tool_item.connect("parent-set", self._tool_item_parent_set)
        return tool_item

    def do_create_menu_item(self):
        """Returns a new MenuItem

        Invoked by UIManager when it needs a MenuItem proxy for a menu.

        This method instantiates and returns a new widget from a class named
        after the factory action's own name.  Class lookup is done via GObject:
        see `TOOL_ITEM_NAME_PATTERN` for the ``__gtype_name__`` this method
        will expect.

        """
        gtype_name = self.MENU_ITEM_NAME_PATTERN % (self.get_name(),)
        menu_item = self._construct(gtype_name)
        #menu_item.connect("parent-set", self._tool_item_parent_set)
        return menu_item

    def _construct(self, gtype_name):
        try:
            gtype = GObject.type_from_name(gtype_name)
        except RuntimeError:
            warn("Cannot construct a new %s: not loaded?" % (gtype_name,),
                 RuntimeWarning)
            return None
        if not gtype.is_a(Gtk.Widget):
            warn("%s is not a Gtk.Widget subclass" % (gtype_name,),
                 RuntimeWarning)
            return None
        widget = gtype.pytype()
        return widget

    def _tool_item_parent_set(self, widget, old_parent):
        parent = widget.get_parent()
        if parent and parent.get_visible():
            widget.show_all()



# This file is part of MyPaint.
# Copyright (C) 2008-2013 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2010-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import division, print_function

from math import ceil
import logging

from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GdkPixbuf
from lib.gibindings import GLib

from lib import helpers
from lib.observable import event
from . import uicolor
from lib.pycompat import xrange

logger = logging.getLogger(__name__)

DRAG_ITEM_NAME = 'text/plain'
DRAG_ITEM_ID = 103
DRAG_TARGETS = [(DRAG_ITEM_NAME, Gtk.TargetFlags.SAME_APP, DRAG_ITEM_ID)]
ITEM_SIZE_DEFAULT = 48


class PixbufList(Gtk.DrawingArea):

    def on_drag_data(self, copy, source_widget, brush_name, target_idx):
        return False

    def drag_begin_cb(self, widget, context):
        widget.drag_insertion_index = None

    def drag_end_cb(self, widget, context):
        widget.drag_insertion_index = None

    # GType naming, for GtkBuilder
    __gtype_name__ = 'PixbufList'

    def __init__(self, itemlist=None,
                 item_w=ITEM_SIZE_DEFAULT,
                 item_h=ITEM_SIZE_DEFAULT,
                 namefunc=None, pixbuffunc=lambda x: x,
                 idfunc=lambda x: str(id(x))):
        super(PixbufList, self).__init__()

        if itemlist is not None:
            self.itemlist = itemlist
        else:
            self.itemlist = []
        self.pixbuffunc = pixbuffunc
        self.namefunc = namefunc
        self.idfunc = idfunc
        self.dragging_allowed = True

        self.pixbuf = None
        self.spacing_outside = 0
        self.border_visible = 2
        self.border_visible_outside_cell = 1
        self.spacing_inside = 0
        self.set_size(item_w, item_h)

        self.selected = None
        self.tooltip_text = None
        self.in_potential_drag = False

        self.connect("draw", self.draw_cb)

        self.connect("button-press-event", self.button_press_cb)
        self.connect("button-release-event", self.button_release_cb)
        self.connect("configure-event", self.configure_event_cb)
        self.connect("motion-notify-event", self.motion_notify_cb)
        self.set_events(
            Gdk.EventMask.BUTTON_PRESS_MASK |
            Gdk.EventMask.BUTTON_RELEASE_MASK |
            Gdk.EventMask.POINTER_MOTION_MASK |
            Gdk.EventMask.EXPOSURE_MASK
        )

        self.realized_once = False
        self.connect("realize", self.on_realize)

        self.drag_highlighted = False
        self.drag_insertion_index = None

        # Fake a nicer style
        style_context = self.get_style_context()
        style_context.add_class(Gtk.STYLE_CLASS_VIEW)

        self.update()

    def on_realize(self, widget):
        if self.realized_once:
            return
        self.realized_once = True
        if self.dragging_allowed:
            # DnD setup.
            self.connect('drag-data-get', self.drag_data_get_cb)
            self.connect('drag-motion', self.drag_motion_cb)
            self.connect('drag-leave', self.drag_leave_cb)
            self.connect('drag-begin', self.drag_begin_cb)
            self.connect('drag-end', self.drag_end_cb)
            self.connect('drag-data-received', self.drag_data_received_cb)
            # Users can drag pixbufs *to* anywhere on a pixbuflist
            # at all times.
            targets_list = [Gtk.TargetEntry.new(*e) for e in DRAG_TARGETS]
            self.drag_dest_set(Gtk.DestDefaults.ALL, targets_list,
                               Gdk.DragAction.MOVE | Gdk.DragAction.COPY)
            # Dragging *from* a list can only happen over a pixbuf:
            # see motion_notify_cb().
            self.drag_source_sensitive = False

    def set_itemlist(self, items):
        self.itemlist = items
        self.update()

    def set_size(self, item_w, item_h):
        self.item_w = item_w
        self.item_h = item_h
        self.thumbnails = {}

    def motion_notify_cb(self, widget, event):
        over_item = False
        if self.point_is_inside(event.x, event.y):
            i = self.index(event.x, event.y)
            over_item = i < len(self.itemlist)
        if over_item:
            if self.namefunc is not None:
                item = self.itemlist[i]
                item_name = self.namefunc(item)
                # Tooltip changing has to happen over two motion-notifys
                # because we want to force the tooltip box to move with the
                # mouse pointer.
                if self.tooltip_text != item_name:
                    self.tooltip_text = item_name
                    self.set_has_tooltip(False)
                    # pop down on the 1st event with this name
                else:
                    self.set_tooltip_text(item_name)
                    # pop up on the 2nd
            if self.dragging_allowed:
                if not self.drag_source_sensitive:
                    targets_list = [Gtk.TargetEntry.new(*e) for e in
                                    DRAG_TARGETS]
                    self.drag_source_set(
                        Gdk.ModifierType.BUTTON1_MASK, targets_list,
                        Gdk.DragAction.COPY | Gdk.DragAction.MOVE)
                    self.drag_source_sensitive = True
        else:
            if self.tooltip_text is not None:
                self.set_has_tooltip(False)
                self.tooltip_text = None
            if self.dragging_allowed and self.drag_source_sensitive:
                if not self.in_potential_drag:
                    # If we haven't crossed the drag threshold yet, don't kill
                    # the potential drag before it starts.
                    self.drag_source_unset()
                    self.drag_source_sensitive = False

    def drag_motion_cb(self, widget, context, x, y, time):
        if not self.dragging_allowed:
            return False
        action = None
        source_widget = Gtk.drag_get_source_widget(context)
        if self is source_widget:
            # Only moves are possible
            action = Gdk.DragAction.MOVE
        else:
            # Dragging from another widget, is always a copy
            action = Gdk.DragAction.COPY
        Gdk.drag_status(context, action, time)
        if not self.drag_highlighted:
            self.drag_highlighted = True
            self.queue_draw()
        if self.drag_highlighted:
            i = self.index(x, y)
            if i != self.drag_insertion_index:
                self.queue_draw()
                self.drag_insertion_index = i

    def drag_leave_cb(self, widget, context, time):
        if widget.drag_highlighted:
            # widget.drag_unhighlight()   # XXX nonfunctional
            widget.drag_highlighted = False
            widget.drag_insertion_index = None
            widget.queue_draw()

    def drag_data_get_cb(self, widget, context, selection, info, time):
        """Gets the selected brush's drag id into `selection` when requested"""
        if info != DRAG_ITEM_ID:
            return False
        item = self.selected
        assert item in self.itemlist
        dragid = self.idfunc(item)
        selection.set_text(dragid, -1)
        logger.debug(
            "drag-data-get: sending type=%r",
            selection.get_data_type(),
        )
        logger.debug("drag-data-get: sending fmt=%r", selection.get_format())
        logger.debug("drag-data-get: sending data=%r len=%r",
                     selection.get_data(), len(selection.get_data()))
        return True

    def drag_data_received_cb(self, widget, context, x, y, selection,
                              info, time):
        if info != DRAG_ITEM_ID:
            return False
        data = selection.get_text()
        data_type = selection.get_data_type()
        fmt = selection.get_format()
        logger.debug("drag-data-received: got type=%r", data_type)
        logger.debug("drag-data-received: got fmt=%r", fmt)
        logger.debug("drag-data-received: got data=%r len=%r", data, len(data))

        # isc always valid, we reject drops at invalid idx
        target_item_idx = self.index(x, y)

        src_widget = Gtk.drag_get_source_widget(context)
        is_copy = context.get_selected_action() == Gdk.DragAction.COPY
        success = self.on_drag_data(is_copy, src_widget, data, target_item_idx)
        context.finish(success, False, time)
        return True

    def update(self, width=None, height=None):
        """
        Redraws the widget from scratch.
        """
        self.total_border = self.border_visible \
            + self.spacing_inside + self.spacing_outside
        self.total_w = self.item_w + 2*self.total_border
        self.total_h = self.item_h + 2*self.total_border

        if width is None:
            if not self.pixbuf:
                return
            width = self.pixbuf.get_width()
            height = self.pixbuf.get_height()
        width = max(width, self.total_w)
        self.tiles_w = max(1, int(width // self.total_w))
        self.tiles_h = max(1, int(ceil(len(self.itemlist) / self.tiles_w)))

        height = self.tiles_h * self.total_h
        # self.set_size_request(-1, -1)
        GLib.idle_add(self.set_size_request, self.total_w, height)

        self.pixbuf = GdkPixbuf.Pixbuf.new(
            GdkPixbuf.Colorspace.RGB, True,
            8, width, height
        )
        self.pixbuf.fill(0xffffff00)  # transparent
        for i, item in enumerate(self.itemlist):
            x = (i % self.tiles_w) * self.total_w
            y = (i // self.tiles_w) * self.total_h
            x += self.total_border
            y += self.total_border

            pixbuf = self.pixbuffunc(item)
            if pixbuf not in self.thumbnails:
                self.thumbnails[pixbuf] = helpers.pixbuf_thumbnail(
                    pixbuf,
                    self.item_w, self.item_h,
                )
            pixbuf = self.thumbnails[pixbuf]
            pixbuf.composite(
                self.pixbuf, x, y, self.item_w, self.item_h, x, y, 1, 1,
                GdkPixbuf.InterpType.BILINEAR, 255)

        self.queue_draw()

    def set_selected(self, item):
        self.selected = item
        self.queue_draw()

    def index(self, x, y):
        x, y = int(x), int(y)
        i = x // self.total_w
        if i >= self.tiles_w:
            i = self.tiles_w - 1
        if i < 0:
            i = 0
        i = i + self.tiles_w * (y // self.total_h)
        if i < 0:
            i = 0
        return i

    def point_is_inside(self, x, y):
        alloc = self.get_allocation()
        w = alloc.width
        h = alloc.height
        return x >= 0 and y >= 0 and x < w and y < h

    def button_press_cb(self, widget, event):
        ex, ey = int(event.x), int(event.y)
        if not self.point_is_inside(ex, ey):
            return False
        i = self.index(ex, ey)
        if i >= len(self.itemlist):
            return
        item = self.itemlist[i]
        if (event.button == 3):
            self.set_selected(item)
            self.item_popup(item)
        else:
            self.set_selected(item)
            self.item_selected(item)
            if self.selected is not None:
                # early exception if drag&drop would break
                assert self.selected in self.itemlist, \
                    'selection failed: the user selected %r " \
                    "by pointing at it, but after calling item_selected() " \
                    "%r is active instead!' % (item, self.selected)
            self.in_potential_drag = True

    @event
    def item_selected(self, item):
        """Event: the user selected an item in the list"""

    @event
    def item_popup(self, item):
        """Event: the user brought up the popup menu for an item in the list"""

    def button_release_cb(self, widget, event):
        self.in_potential_drag = False

    def configure_event_cb(self, widget, size):
        if self.pixbuf and self.pixbuf.get_width() == size.width:
            if self.pixbuf.get_height() == size.height:
                return
        self.update(size.width, size.height)

    def draw_cb(self, widget, cr):
        # Paint the base color, and the list's pixbuf.
        state_flags = widget.get_state_flags()
        style_context = widget.get_style_context()
        style_context.save()
        bg_color_gdk = style_context.get_background_color(state_flags)
        bg_color = uicolor.from_gdk_rgba(bg_color_gdk)
        cr.set_source_rgb(*bg_color.get_rgb())
        cr.paint()
        Gdk.cairo_set_source_pixbuf(cr, self.pixbuf, 0, 0)
        cr.paint()
        # border colors
        gdkrgba = style_context.get_background_color(
            state_flags | Gtk.StateFlags.SELECTED)
        selected_color = uicolor.from_gdk_rgba(gdkrgba)
        gdkrgba = style_context.get_background_color(
            state_flags | Gtk.StateFlags.NORMAL)
        insertion_color = uicolor.from_gdk_rgba(gdkrgba)
        style_context.restore()
        # Draw borders
        last_i = len(self.itemlist) - 1
        for i, b in enumerate(self.itemlist):
            rect_color = None
            if b is self.selected:
                rect_color = selected_color
            elif self.drag_insertion_index is not None:
                if i == self.drag_insertion_index \
                        or (i == last_i and self.drag_insertion_index > i):
                    rect_color = insertion_color
            if rect_color is None:
                continue
            x = (i % self.tiles_w) * self.total_w
            y = (i // self.tiles_w) * self.total_h
            w = self.total_w
            h = self.total_h

            def shrink(pixels, x, y, w, h):
                x += pixels
                y += pixels
                w -= 2*pixels
                h -= 2*pixels
                return (x, y, w, h)
            x, y, w, h = shrink(self.spacing_outside, x, y, w, h)
            for j in xrange(self.border_visible_outside_cell):
                x, y, w, h = shrink(-1, x, y, w, h)
            max_j = self.border_visible + self.border_visible_outside_cell
            for j in xrange(max_j):
                cr.set_source_rgb(*rect_color.get_rgb())
                cr.rectangle(x, y, w-1, h-1)   # FIXME: check pixel alignment
                cr.stroke()
                x, y, w, h = shrink(1, x, y, w, h)
        return True


if __name__ == '__main__':
    logging.basicConfig()
    win = Gtk.Window()
    win.set_title("pixbuflist test")
    test_list = PixbufList()
    win.add(test_list)
    test_list.set_size_request(256, 128)
    win.connect("destroy", lambda *a: Gtk.main_quit())
    win.show_all()
    Gtk.main()



# This file is part of MyPaint.
# Copyright (C) 2017 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Abstract definitions of how to render layer trees.

Rendering is performed by RootLayerStack objects: see lib.layer.tree
and its "render_*()" methods.

"""

# Imports:

import abc


# Public constants:

class Opcode:
    """Opcode values for Renderable.get_render_ops()."""

    #: Composite some data into the backdrop.
    #: Format: (COMPOSITE, <TileCompositable>, modenum, opacity)
    COMPOSITE = 1

    #: Blit some data into the backdrop.
    #: Format: (BLIT, <TileBlittable>, None, None)
    BLIT = 2

    #: Push an empty tile onto the isolation stack.
    #: This creates a new isolated backdrop for later ops.
    #: Format: (PUSH, None, None, opacity)
    PUSH = 3

    #: Pop the isolation stack & composite the removed tile.
    #: The removed backdrop is composited onto the previous one.
    #: Format: (POP, None, modenum, opacity)
    POP = 4


# Classes and interfaces:

class Spec (object):
    """Selection criteria for Renderable.get_render_ops()."""

    def __init__(self, **kwargs):
        """Initialize with optional field info from **kwargs."""
        super(Spec, self).__init__()

        #: Limitation: render *only* these layers, if specified.
        #: If a child layer is present here,
        #: its entire parent chain must be present too.
        #: Default: None.
        #: Type: set() of Renderables.
        self.layers = kwargs.get("layers", None)

        #: The currently active layer.
        #: Renderables may choose to render themselves differently
        #: if they are the current layer. Other flags may apply too.
        #: Default: None.
        self.current = kwargs.get("current", None)

        #: Flag: the layer in "current" is being shown in solo mode.
        #: The "layers" field is expected to hold all the parents.
        #: Default: False.
        self.solo = kwargs.get("solo", False)

        #: Flag: the layer in "current" is being shown as a preview effect.
        #: The "layers" field is expected to hold all the parents.
        #: Default: False.
        self.previewing = kwargs.get("previewing", False)

        #: Effect: a Renderable to be rendered on top of everything.
        #: Default: None.
        self.global_overlay = kwargs.get("global_overlay", None)

        #: Effect: a Renderable to be rendered on top of "current".
        #: See also "global_overlay".
        #: Default: None.
        self.current_overlay = kwargs.get("current_overlay", None)

        #: Tristate flag: override for the the visible state of the
        #: builtin background layer. Leave this one unset if possible.
        #: Default: None (use the background layer's own visibility flag)
        flagval = kwargs.get("background", None)
        if flagval is not None:
            flagval = bool(flagval)
        self.background = flagval

    def cacheable(self):
        """True if render methods can usefully cache output for this spec."""
        result = (self.current_overlay is None)
        result &= (self.global_overlay is None)
        result &= not bool(self.solo)
        result &= not bool(self.previewing)
        result &= (self.layers is None)
        result &= (self.background is None)
        return result


class Renderable:
    """Abstract interface for elements that can be rendered."""

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def get_render_ops(self, spec):
        """Returns a flat sequence of rendering ops.

        Implementations of this function are expected
        to perform all decisions about what to render,
        and it is reasonable to make quite complex decisions here.
        This method will be called on the entire layer stack
        before any tiles are accessed and composited.
        It's expected that it will visit descendents recursively
        and serve up a small program in rendering order.

        :param lib.layer.rendering.Spec spec: what to render.
        :rtype: sequence
        :returns: sequence of ops

        The "spec" parameter works pretty much as a query.
        The tree of renderables being searched/visited
        may choose to return different things
        based on what it contains.
        See: lib.layer.rendering.Spec.

        The returned sequence of ops needs to be completely flat.
        Each op is a tuple, [(OPCODE, DATA, MODE, OPACITY), ...]

        OPCODE is an Opcode constant.
        The DATA, MODE, and OPACITY members will vary
        according to the specific operation.
        See lib.layer.rendering.Opcode for details.

        """
